import inspect
import itertools
import os
import re
import subprocess
import sys
import sysconfig
import textwrap
from unittest.mock import ANY

from memray import AllocatorType
from memray import FileFormat
from memray import FileReader
from memray._memray import RecordWriterTestHarness


def get_sample_line_number_tables():
    callers = []

    def save_caller():
        stack = inspect.stack()
        caller = stack[1].frame
        linetable = RecordWriterTestHarness.get_linetable(caller.f_code)

        callers.append(
            {
                "lasti": RecordWriterTestHarness.get_lasti(caller),
                "linetable": linetable,
            }
        )

    def foo():
        x = 10
        y = 20
        _ = x + y
        save_caller()
        return

    def bar():
        x = []
        for _ in range(100):
            x.append(x)

        save_caller()

        return None

    def baz():
        yield
        save_caller()
        return "foo"

    foo()
    bar()
    list(baz())
    return callers


def parse_capture_file(output_file):
    output = subprocess.check_output(
        [sys.executable, "-m", "memray", "parse", output_file],
        encoding="utf-8",
    )
    parsed = output.splitlines()
    header, *rest = parsed

    assert header.startswith("HEADER ")
    header_fields = list(re.findall(r"(\S+)=(.*?)\s*(?=\S+=|$)", header))
    return header_fields, rest


def sort_runs_of_same_record_type(records):
    sorted_records = []
    for _, group in itertools.groupby(records, lambda line: line.split(" ", 1)[0]):
        sorted_records.extend(sorted(group))
    return sorted_records


def test_write_basic_records(tmp_path):
    """Test writing basic records to a file."""
    # GIVEN
    output_file = tmp_path / "test.memray"

    # WHEN
    writer = RecordWriterTestHarness(
        str(output_file),
        native_traces=True,
        trace_python_allocators=True,
        file_format=FileFormat.ALL_ALLOCATIONS,
        main_tid=10,
        skipped_frames=15,
        libdest='/some/path with spaces and a ":" too',
        site_packages=["/some/path", '/"another"/path too'],
        sys_path=["/some/path", '/"another"/path too'],
    )

    assert writer.write_memory_record(1757101101880, 1024)
    assert writer.write_memory_record(1757101102880, 2048)

    assert writer.write_thread_record(1, "MainThread")
    assert writer.write_thread_record(2, "Other Thread")

    assert writer.write_code_object(1, "a", "a.py", b"", 1)
    assert writer.write_code_object(2, "b", "b.py", b"", 11)
    assert writer.write_code_object(3, "c", "c.py", b"", 21)

    assert writer.write_frame_push(1, 1, 12, True)
    assert writer.write_frame_push(1, 2, 34, True)
    assert writer.write_frame_push(1, 3, 56, True)

    assert writer.write_frame_pop(1, 0)
    assert writer.write_frame_pop(1, 1)
    assert writer.write_frame_pop(1, 16)
    assert writer.write_frame_pop(1, 17)

    assert writer.write_unresolved_native_frame(0xDEADBEEF, 1)
    assert writer.write_unresolved_native_frame(0xFEEDCAFE, 2)

    assert writer.write_allocation_record(1, 0x1000, 1024, AllocatorType.MALLOC, 1)
    assert writer.write_allocation_record(1, 0x2000, 2048, AllocatorType.PYMALLOC_FREE)
    assert writer.write_allocation_record(1, 0x3000, 4096, AllocatorType.POSIX_MEMALIGN)

    assert writer.write_allocation_record(2, 0x4000, 4096, AllocatorType.MALLOC, 2)

    assert writer.write_mappings(
        [
            {
                "filename": "/usr/lib/libc.so.6",
                "addr": 0xF1234567,
                "segments": [
                    {"vaddr": 0x12345678, "memsz": 0x1000},
                    {"vaddr": 0xF2345678, "memsz": 0x2000},
                ],
            }
        ]
    )

    assert writer.write_trailer()

    # THEN
    header_fields, records = parse_capture_file(output_file)
    allocator = (
        "mimalloc" if sysconfig.get_config_var("Py_GIL_DISABLED") else "pymalloc"
    )

    assert header_fields == [
        ("magic", "memray"),
        ("version", "13"),
        ("python_version", f"{sys.hexversion:08x}"),
        ("native_traces", "true"),
        ("file_format", "ALL_ALLOCATIONS"),
        ("n_allocations", "0"),
        ("n_frames", "0"),
        ("start_time", ANY),
        ("end_time", ANY),
        ("pid", f"{os.getpid()}"),
        ("main_tid", "10"),
        ("skipped_frames_on_main_tid", "15"),
        ("command_line", "memray test harness"),
        ("python_allocator", allocator),
        ("trace_python_allocators", "true"),
        ("track_object_lifetimes", "false"),
        ("libdest", '/some/path with spaces and a ":" too'),
        ("site_packages", '[/some/path, /"another"/path too]'),
        ("sys.path", '[/some/path, /"another"/path too]'),
    ]

    expected_parse_output = """
        MEMORY_RECORD time=1757101101880 memory=1024
        MEMORY_RECORD time=1757101102880 memory=2048
        CONTEXT_SWITCH tid=1
        THREAD MainThread
        CONTEXT_SWITCH tid=2
        THREAD Other Thread
        CODE_OBJECT code_id=1 function_name=a filename=a.py firstlineno=1 linetable_size=0
        CODE_OBJECT code_id=2 function_name=b filename=b.py firstlineno=11 linetable_size=0
        CODE_OBJECT code_id=3 function_name=c filename=c.py firstlineno=21 linetable_size=0
        CONTEXT_SWITCH tid=1
        FRAME_PUSH code_object_id=1 instruction_offset=12 is_entry_frame=1
        FRAME_PUSH code_object_id=2 instruction_offset=34 is_entry_frame=1
        FRAME_PUSH code_object_id=3 instruction_offset=56 is_entry_frame=1
        FRAME_POP count=1
        FRAME_POP count=16
        FRAME_POP count=16
        FRAME_POP count=1
        NATIVE_FRAME_ID ip=0xdeadbeef index=1
        NATIVE_FRAME_ID ip=0xfeedcafe index=2
        ALLOCATION address=0x1000 size=1024 allocator=malloc native_frame_id=1
        ALLOCATION address=0x2000 size=0 allocator=pymalloc_free native_frame_id=0
        ALLOCATION address=0x3000 size=4096 allocator=posix_memalign native_frame_id=0
        CONTEXT_SWITCH tid=2
        ALLOCATION address=0x4000 size=4096 allocator=malloc native_frame_id=2
        MEMORY_MAP_START
        SEGMENT_HEADER filename=/usr/lib/libc.so.6 num_segments=2 addr=0xf1234567
        SEGMENT 0x12345678 1000
        SEGMENT 0xf2345678 2000
        TRAILER
    """

    expected_records = textwrap.dedent(expected_parse_output).strip().splitlines()
    assert records == expected_records


def test_snapshot_allocations_preserve_capture_order(tmp_path):
    output_file = tmp_path / "ordered.memray"
    writer = RecordWriterTestHarness(str(output_file))

    assert writer.write_code_object(1, "location", "location.py", b"", 1)

    assert writer.write_frame_push(1, 1, 0, True)
    assert writer.write_allocation_record(1, 0x1000, 1024, AllocatorType.MALLOC)

    assert writer.write_frame_push(2, 1, 0, True)
    assert writer.write_allocation_record(2, 0x2000, 2048, AllocatorType.MALLOC)

    assert writer.write_allocation_record(1, 0x3000, 512, AllocatorType.MALLOC)
    assert writer.write_trailer()

    with FileReader(output_file) as reader:
        allocations = list(
            reader.get_high_watermark_allocation_records(merge_threads=False)
        )

    assert [record.tid for record in allocations] == [1, 2]
    assert [(record.size, record.n_allocations) for record in allocations] == [
        (1536, 2),
        (2048, 1),
    ]


def test_write_aggregated_records(tmp_path):
    """Test writing aggregated records to a file."""
    # GIVEN
    output_file = tmp_path / "test_aggregated.memray"

    # WHEN
    writer = RecordWriterTestHarness(
        str(output_file),
        native_traces=False,
        trace_python_allocators=False,
        file_format=FileFormat.AGGREGATED_ALLOCATIONS,
    )

    assert writer.write_memory_record(1757105244685, 10000)

    assert writer.write_code_object(10, "a", "a.py", b"", 1)
    assert writer.write_code_object(20, "b", "b.py", b"", 11)
    assert writer.write_code_object(30, "c", "c.py", b"", 21)

    assert writer.write_frame_push(1, 1, 12, True)
    assert writer.write_frame_push(1, 2, 34, True)
    assert writer.write_allocation_record(1, 0x1000, 1024, AllocatorType.MALLOC)

    assert writer.write_frame_push(2, 3, 56, True)
    assert writer.write_allocation_record(2, 0x2000, 2048, AllocatorType.MALLOC)

    assert writer.write_memory_record(1757105245685, 20000)

    assert writer.write_trailer()

    # THEN
    header_fields, records = parse_capture_file(output_file)
    allocator = (
        "mimalloc" if sysconfig.get_config_var("Py_GIL_DISABLED") else "pymalloc"
    )

    assert header_fields == [
        ("magic", "memray"),
        ("version", "13"),
        ("python_version", f"{sys.hexversion:08x}"),
        ("native_traces", "false"),
        ("file_format", "AGGREGATED_ALLOCATIONS"),
        ("n_allocations", "0"),
        ("n_frames", "0"),
        ("start_time", ANY),
        ("end_time", ANY),
        ("pid", f"{os.getpid()}"),
        ("main_tid", "1"),
        ("skipped_frames_on_main_tid", "0"),
        ("command_line", "memray test harness"),
        ("python_allocator", allocator),
        ("trace_python_allocators", "false"),
        ("track_object_lifetimes", "false"),
        ("libdest", ""),
        ("site_packages", "[]"),
        ("sys.path", "[]"),
    ]

    records = sort_runs_of_same_record_type(records)
    expected_parse_output = """
        MEMORY_SNAPSHOT time=1757105244685 rss=10000 heap=0
        MEMORY_SNAPSHOT time=1757105245685 rss=20000 heap=3072
        CODE_OBJECT code_id=10 function_name=a filename=a.py firstlineno=1 linetable_size=0
        CODE_OBJECT code_id=20 function_name=b filename=b.py firstlineno=11 linetable_size=0
        CODE_OBJECT code_id=30 function_name=c filename=c.py firstlineno=21 linetable_size=0
        PYTHON_FRAME_INDEX frame_id=0 code_object_id=1 instruction_offset=12 is_entry_frame=1
        PYTHON_FRAME_INDEX frame_id=1 code_object_id=2 instruction_offset=34 is_entry_frame=1
        PYTHON_FRAME_INDEX frame_id=2 code_object_id=3 instruction_offset=56 is_entry_frame=1
        PYTHON_TRACE_INDEX frame_id=0 parent_index=0
        PYTHON_TRACE_INDEX frame_id=1 parent_index=1
        PYTHON_TRACE_INDEX frame_id=2 parent_index=0
        AGGREGATED_ALLOCATION tid=1 allocator=malloc native_frame_id=0 python_frame_id=2 native_segment_generation=0 n_allocations_in_high_water_mark=1 n_allocations_leaked=1 bytes_in_high_water_mark=1024 bytes_leaked=1024
        AGGREGATED_ALLOCATION tid=2 allocator=malloc native_frame_id=0 python_frame_id=3 native_segment_generation=0 n_allocations_in_high_water_mark=1 n_allocations_leaked=1 bytes_in_high_water_mark=2048 bytes_leaked=2048
        AGGREGATED_TRAILER
    """  # noqa: E501
    expected_records = sort_runs_of_same_record_type(
        textwrap.dedent(expected_parse_output).strip().splitlines()
    )

    assert records == expected_records


def test_decoding_line_numbers(tmp_path):
    # GIVEN
    funcs = get_sample_line_number_tables()
    output_file = tmp_path / "test.memray"

    # WHEN
    writer = RecordWriterTestHarness(str(output_file))

    assert writer.write_code_object(1, "foo", "foo a.py", funcs[0]["linetable"], 1)
    assert writer.write_code_object(2, "bar", "bar b.py", funcs[1]["linetable"], 11)
    assert writer.write_code_object(3, "baz", "baz c.py", funcs[2]["linetable"], 21)

    assert writer.write_frame_push(1, 1, funcs[0]["lasti"], True)
    assert writer.write_frame_push(1, 2, funcs[1]["lasti"], True)
    assert writer.write_frame_push(1, 3, funcs[2]["lasti"], True)

    assert writer.write_allocation_record(1, 0x1000, 1024, AllocatorType.MALLOC)

    # THEN
    with FileReader(output_file) as reader:
        allocations = list(reader.get_allocation_records())
        assert len(allocations) == 1
        assert allocations[0].stack_trace() == [
            ("baz", "baz c.py", 23),
            ("bar", "bar b.py", 16),
            ("foo", "foo a.py", 5),
        ]


def test_write_object_tracking_records(tmp_path):
    """Test writing object tracking records to a file."""
    # GIVEN
    output_file = tmp_path / "test_object_tracking.memray"

    # WHEN
    writer = RecordWriterTestHarness(
        str(output_file),
        native_traces=True,
        trace_python_allocators=False,
        track_object_lifetimes=True,
        file_format=FileFormat.ALL_ALLOCATIONS,
        main_tid=1,
    )

    # Write object tracking records
    assert writer.write_object_record(1, 0x12345678, True, 0)  # Object created
    assert writer.write_object_record(
        1, 0x87654321, True, 1
    )  # Object created with native frame
    assert writer.write_object_record(1, 0x12345678, False, 0)  # Object destroyed
    assert writer.write_object_record(
        1, 0x87654321, False, 1
    )  # Object destroyed with native frame

    assert writer.write_trailer()

    # THEN
    header_fields, records = parse_capture_file(output_file)

    # Check that track_object_lifetimes is in the header
    header_dict = dict(header_fields)
    assert header_dict.get("track_object_lifetimes") == "true"

    expected_parse_output = """
    CONTEXT_SWITCH tid=1
    OBJECT_RECORD address=0x12345678 native_frame_id=0
    OBJECT_RECORD address=0x87654320 native_frame_id=1
    OBJECT_RECORD address=0x12345678 native_frame_id=0
    OBJECT_RECORD address=0x87654320 native_frame_id=0
    TRAILER
    """

    expected_records = textwrap.dedent(expected_parse_output).strip().splitlines()
    assert records == expected_records


def test_write_object_tracking_records_aggregated(tmp_path):
    # GIVEN
    output_file = tmp_path / "test_object_tracking_aggregated.memray"

    # WHEN
    writer = RecordWriterTestHarness(
        str(output_file),
        native_traces=True,
        trace_python_allocators=False,
        track_object_lifetimes=True,
        file_format=FileFormat.AGGREGATED_ALLOCATIONS,
        main_tid=1,
    )

    # Write object tracking records - create and destroy some objects
    assert writer.write_object_record(1, 0x12345678, True, 0)  # Object created
    assert writer.write_object_record(
        1, 0x87654321, True, 1
    )  # Object created with native frame
    assert writer.write_object_record(1, 0xAABBCCDD, True, 2)  # Another object created
    assert writer.write_object_record(1, 0x12345678, False, 0)  # First object destroyed
    # 0x87654321 and 0xAABBCCDD survive

    assert writer.write_trailer()

    # THEN
    header_fields, records = parse_capture_file(output_file)

    # Check that track_object_lifetimes is in the header
    header_dict = dict(header_fields)
    assert header_dict.get("track_object_lifetimes") == "true"
    assert header_dict.get("file_format") == "AGGREGATED_ALLOCATIONS"

    expected_parse_output = """
    SURVIVING_OBJECT address=0x0xaabbccd8 native_frame_id=2
    SURVIVING_OBJECT address=0x0x87654320 native_frame_id=1
    AGGREGATED_TRAILER
    """

    records = sort_runs_of_same_record_type(records)
    expected_records = sort_runs_of_same_record_type(
        textwrap.dedent(expected_parse_output).strip().splitlines()
    )

    assert records == expected_records



import functools
from collections import Counter

# DO NOT CHANGE
FIRST_COUNTER_RANGE = 500
SECOND_COUNTER_RANGE = 1000
# DO NOT CHANGE


class Algorithms:
    def __init__(self, inc: int):
        self.inc = inc
        self.factorial_plus = functools.cache(self._uncached_factorial_plus)

    def _uncached_factorial_plus(self, n: int) -> int:
        return n * self.factorial_plus(n - 1) + self.inc if n > 1 else 1 + self.inc


def generate_factorial_plus_last_digit(plus_range: int, factorial_range: int):
    for i in range(plus_range):
        A = Algorithms(i)
        for j in range(factorial_range):
            yield A.factorial_plus(j) % 10


def compare_counts_different_factorials():
    counts_500 = Counter(
        generate_factorial_plus_last_digit(FIRST_COUNTER_RANGE, FIRST_COUNTER_RANGE)
    )
    counts_1000 = Counter(
        generate_factorial_plus_last_digit(SECOND_COUNTER_RANGE, SECOND_COUNTER_RANGE)
    )
    print(counts_500.most_common())
    print(counts_1000.most_common())


if __name__ == "__main__":
    compare_counts_different_factorials()



"""Script for testing the performance of json parsing and serialization.

This will dump/load several real world-representative objects a few
thousand times. The methodology below was chosen for was chosen to be similar
to real-world scenarios which operate on single objects at a time.
"""

# Python imports
import json
import random
import sys

# Local imports
import pyperf
from memray_helper import get_tracker


DICT = {
    "ads_flags": 0,
    "age": 18,
    "bulletin_count": 0,
    "comment_count": 0,
    "country": "BR",
    "encrypted_id": "G9urXXAJwjE",
    "favorite_count": 9,
    "first_name": "",
    "flags": 412317970704,
    "friend_count": 0,
    "gender": "m",
    "gender_for_display": "Male",
    "id": 302935349,
    "is_custom_profile_icon": 0,
    "last_name": "",
    "locale_preference": "pt_BR",
    "member": 0,
    "tags": ["a", "b", "c", "d", "e", "f", "g"],
    "profile_foo_id": 827119638,
    "secure_encrypted_id": "Z_xxx2dYx3t4YAdnmfgyKw",
    "session_number": 2,
    "signup_id": "201-19225-223",
    "status": "A",
    "theme": 1,
    "time_created": 1225237014,
    "time_updated": 1233134493,
    "unread_message_count": 0,
    "user_group": "0",
    "username": "collinwinter",
    "play_count": 9,
    "view_count": 7,
    "zip": "",
}

TUPLE = (
    [
        265867233,
        265868503,
        265252341,
        265243910,
        265879514,
        266219766,
        266021701,
        265843726,
        265592821,
        265246784,
        265853180,
        45526486,
        265463699,
        265848143,
        265863062,
        265392591,
        265877490,
        265823665,
        265828884,
        265753032,
    ],
    60,
)


def mutate_dict(orig_dict, random_source):
    new_dict = dict(orig_dict)
    for key, value in new_dict.items():
        rand_val = random_source.random() * sys.maxsize
        if isinstance(key, (int, bytes, str)):
            new_dict[key] = type(key)(rand_val)
    return new_dict


random_source = random.Random(5)  # Fixed seed.
DICT_GROUP = [mutate_dict(DICT, random_source) for _ in range(3)]


def bench_json_loads(loops, objs):
    with get_tracker():
        t0 = pyperf.perf_counter()
        for _ in range(loops):
            for obj in objs:
                # 20 loads
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
                json.loads(obj)
        dt = pyperf.perf_counter() - t0
    return dt


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata["description"] = "Benchmark json.loads()"

    json_dict = json.dumps(DICT)
    json_tuple = json.dumps(TUPLE)
    json_dict_group = json.dumps(DICT_GROUP)
    objs = (json_dict, json_tuple, json_dict_group)

    runner.bench_time_func("json_loads", bench_json_loads, objs)



import argparse
import sys
from pathlib import Path
from unittest.mock import patch

import pytest

from memray import FileDestination
from memray import SocketDestination
from memray.commands import main
from memray.commands.flamegraph import FlamegraphCommand
from memray.commands.run import RunCommand
from memray.commands.stats import StatsCommand
from memray.commands.summary import SummaryCommand
from memray.commands.table import TableCommand
from memray.commands.transform import TransformCommand
from memray.commands.tree import TreeCommand


def test_no_args_passed(capsys):
    with pytest.raises(SystemExit):
        main([])

    captured = capsys.readouterr()
    assert "usage:" in captured.err
    assert "error: the following arguments are required:" in captured.err


@patch.object(RunCommand, "validate_target_file")
@patch("memray.commands.run.Tracker")
@patch("memray.commands.run.runpy")
@patch("memray.commands.run.os.getpid")
class TestRunSubCommand:
    def test_run_without_arguments(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock, capsys
    ):
        with pytest.raises(SystemExit):
            main(["run"])

        captured = capsys.readouterr()
        assert "usage: memray run [-m module | -c cmd | file] [args]" in captured.err

    def test_run_default_output(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock
    ):
        getpid_mock.return_value = 0
        assert 0 == main(["run", "-m", "foobar"])
        runpy_mock.run_module.assert_called_with(
            "foobar", run_name="__main__", alter_sys=True
        )
        tracker_mock.assert_called_with(
            destination=FileDestination("memray-foobar.0.bin", overwrite=False),
            native_traces=False,
        )

    def test_run_with_native_mode(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock
    ):
        getpid_mock.return_value = 0
        assert 0 == main(["run", "--native", "-m", "foobar"])
        runpy_mock.run_module.assert_called_with(
            "foobar", run_name="__main__", alter_sys=True
        )
        tracker_mock.assert_called_with(
            destination=FileDestination("memray-foobar.0.bin", overwrite=False),
            native_traces=True,
        )

    def test_run_with_pymalloc_tracing(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock
    ):
        getpid_mock.return_value = 0
        assert 0 == main(["run", "--trace-python-allocators", "-m", "foobar"])
        runpy_mock.run_module.assert_called_with(
            "foobar", run_name="__main__", alter_sys=True
        )
        tracker_mock.assert_called_with(
            destination=FileDestination("memray-foobar.0.bin", overwrite=False),
            native_traces=False,
            trace_python_allocators=True,
        )

    def test_run_override_output(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock
    ):
        assert 0 == main(["run", "--output", "my_output", "-m", "foobar"])
        runpy_mock.run_module.assert_called_with(
            "foobar", run_name="__main__", alter_sys=True
        )
        tracker_mock.assert_called_with(
            destination=FileDestination("my_output", overwrite=False),
            native_traces=False,
        )

    def test_run_overwrite_output_file(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock
    ):
        assert 0 == main(["run", "-o", "my_output", "-f", "-m", "foobar"])
        runpy_mock.run_module.assert_called_with(
            "foobar", run_name="__main__", alter_sys=True
        )
        tracker_mock.assert_called_with(
            destination=FileDestination("my_output", overwrite=True),
            native_traces=False,
        )

    def test_run_buffered_file_io(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock
    ):
        assert 0 == main(
            ["run", "-o", "my_output", "--buffered-file-io", "-m", "foobar"]
        )
        runpy_mock.run_module.assert_called_with(
            "foobar", run_name="__main__", alter_sys=True
        )
        tracker_mock.assert_called_with(
            destination=FileDestination("my_output", overwrite=False, buffered=True),
            native_traces=False,
        )

    def test_run_buffered_file_io_rejected_with_live(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock, capsys
    ):
        with pytest.raises(SystemExit):
            main(["run", "--buffered-file-io", "--live", "-m", "foobar"])
        captured = capsys.readouterr()
        assert "--buffered-file-io cannot be used with --live" in captured.err

    def test_run_buffered_file_io_rejected_with_live_remote(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock, capsys
    ):
        with pytest.raises(SystemExit):
            main(["run", "--buffered-file-io", "--live-remote", "-m", "foobar"])
        captured = capsys.readouterr()
        assert "--buffered-file-io cannot be used with --live" in captured.err

    def test_run_module(self, getpid_mock, runpy_mock, tracker_mock, validate_mock):
        assert 0 == main(["run", "-m", "foobar"])
        runpy_mock.run_module.assert_called_with(
            "foobar", run_name="__main__", alter_sys=True
        )

    def test_run_cmd_is_validated(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock
    ):
        with patch.object(RunCommand, "validate_target_file"):
            assert 0 == main(["run", "-c", "x = [i for i in range(10)]"])
            with pytest.raises(SyntaxError):
                main(["run", "-c", "[i for i in range(10)"])

    def test_run_cmd(self, getpid_mock, runpy_mock, tracker_mock, validate_mock):
        with patch("memray.commands.run.exec") as mock_exec:
            assert 0 == main(["run", "-c", "x = 10; y = abs(-10)"])
            assert not runpy_mock.called
            mock_exec.assert_called_with(
                "x = 10; y = abs(-10)", {"__name__": "__main__"}
            )

    def test_run_file(self, getpid_mock, runpy_mock, tracker_mock, validate_mock):
        with patch.object(RunCommand, "validate_target_file"):
            assert 0 == main(["run", "foobar.py", "arg1", "arg2"])
        runpy_mock.run_path.assert_called_with("foobar.py", run_name="__main__")

    def test_run_relative_file(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock
    ):
        getpid_mock.return_value = 0
        with patch.object(RunCommand, "validate_target_file"):
            assert 0 == main(["run", "./directory/foobar.py", "arg1", "arg2"])
        runpy_mock.run_path.assert_called_with(
            "./directory/foobar.py",
            run_name="__main__",
        )
        tracker_mock.assert_called_with(
            destination=FileDestination(
                "./directory/memray-foobar.py.0.bin", overwrite=False
            ),
            native_traces=False,
        )

    @patch("memray.commands.run.subprocess.Popen")
    @patch("memray.commands.run.LiveCommand")
    def test_run_with_live(
        self,
        live_command_mock,
        popen_mock,
        getpid_mock,
        runpy_mock,
        tracker_mock,
        validate_mock,
    ):
        getpid_mock.return_value = 0
        popen_mock().__enter__().returncode = 0
        with patch("memray.commands.run._get_free_port", return_value=1234):
            assert 0 == main(["run", "--live", "./directory/foobar.py", "arg1", "arg2"])
        popen_mock.assert_called_with(
            [
                sys.executable,
                "-c",
                "from memray.commands.run import _child_process;"
                "_child_process(1234,False,False,False,False,False,"
                "'./directory/foobar.py',['arg1', 'arg2'])",
            ],
            stderr=-1,
            stdout=-3,
            text=True,
        )
        live_command_mock().start_live_interface.assert_called_with(
            1234,
            cmdline_override="./directory/foobar.py arg1 arg2",
        )

    @patch("memray.commands.run.subprocess.Popen")
    @patch("memray.commands.run.LiveCommand")
    def test_run_with_live_and_trace_python_allocators(
        self,
        live_command_mock,
        popen_mock,
        getpid_mock,
        runpy_mock,
        tracker_mock,
        validate_mock,
    ):
        getpid_mock.return_value = 0
        popen_mock().__enter__().returncode = 0
        with patch("memray.commands.run._get_free_port", return_value=1234):
            assert 0 == main(
                [
                    "run",
                    "--live",
                    "--trace-python-allocators",
                    "./directory/foobar.py",
                    "arg1",
                    "arg2",
                ]
            )
        popen_mock.assert_called_with(
            [
                sys.executable,
                "-c",
                "from memray.commands.run import _child_process;"
                "_child_process(1234,False,True,False,False,False,"
                "'./directory/foobar.py',['arg1', 'arg2'])",
            ],
            stderr=-1,
            stdout=-3,
            text=True,
        )
        live_command_mock().start_live_interface.assert_called_with(
            1234,
            cmdline_override="./directory/foobar.py arg1 arg2",
        )

    def test_run_with_live_remote(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock
    ):
        getpid_mock.return_value = 0
        with patch("memray.commands.run._get_free_port", return_value=1234):
            assert 0 == main(
                ["run", "--live-remote", "./directory/foobar.py", "arg1", "arg2"]
            )
        runpy_mock.run_path.assert_called_with(
            "./directory/foobar.py",
            run_name="__main__",
        )
        tracker_mock.assert_called_with(
            destination=SocketDestination(server_port=1234, address="127.0.0.1"),
            native_traces=False,
        )

    def test_run_with_live_remote_and_live_port(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock
    ):
        getpid_mock.return_value = 0
        assert 0 == main(
            [
                "run",
                "--live-remote",
                "--live-port=1111",
                "./directory/foobar.py",
                "arg1",
                "arg2",
            ]
        )
        runpy_mock.run_path.assert_called_with(
            "./directory/foobar.py",
            run_name="__main__",
        )
        tracker_mock.assert_called_with(
            destination=SocketDestination(server_port=1111, address="127.0.0.1"),
            native_traces=False,
        )

    def test_run_with_live_port_but_not_live_remote(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock, capsys
    ):
        with pytest.raises(SystemExit):
            main(["run", "--live-port", "1234", "./directory/foobar.py"])

        captured = capsys.readouterr()
        assert "The --live-port argument requires --live-remote" in captured.err

    def test_run_with_follow_fork(
        self,
        getpid_mock,
        runpy_mock,
        tracker_mock,
        validate_mock,
    ):
        getpid_mock.return_value = 0
        assert 0 == main(["run", "--follow-fork", "-m", "foobar"])
        runpy_mock.run_module.assert_called_with(
            "foobar", run_name="__main__", alter_sys=True
        )
        tracker_mock.assert_called_with(
            destination=FileDestination("memray-foobar.0.bin", overwrite=False),
            native_traces=False,
            follow_fork=True,
        )

    def test_run_with_follow_fork_and_live_mode(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock, capsys
    ):
        with pytest.raises(SystemExit):
            main(["run", "--live", "--follow-fork", "./directory/foobar.py"])

        captured = capsys.readouterr()
        assert "--follow-fork cannot be used with" in captured.err

    def test_run_with_follow_fork_and_live_remote_mode(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock, capsys
    ):
        with pytest.raises(SystemExit):
            main(["run", "--live-remote", "--follow-fork", "./directory/foobar.py"])

        captured = capsys.readouterr()
        assert "--follow-fork cannot be used with" in captured.err

    def test_run_with_trace_python_allocators_and_live_remote_mode(
        self, getpid_mock, runpy_mock, tracker_mock, validate_mock, capsys
    ):
        getpid_mock.return_value = 0
        with patch("memray.commands.run._get_free_port", return_value=1234):
            assert 0 == main(
                [
                    "run",
                    "--live-remote",
                    "--trace-python-allocators",
                    "./directory/foobar.py",
                    "arg1",
                    "arg2",
                ]
            )
        runpy_mock.run_path.assert_called_with(
            "./directory/foobar.py",
            run_name="__main__",
        )
        tracker_mock.assert_called_with(
            destination=SocketDestination(server_port=1234, address="127.0.0.1"),
            native_traces=False,
            trace_python_allocators=True,
        )


class TestFlamegraphSubCommand:
    @staticmethod
    def get_prepared_parser():
        parser = argparse.ArgumentParser()
        command = FlamegraphCommand()
        command.prepare_parser(parser)

        return command, parser

    def test_parser_rejects_no_arguments(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args([])

    def test_parser_rejects_when_no_results_provided(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args(["--output", "output.html"])

    def test_parser_accepts_single_argument(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output is None
        assert namespace.show_memory_leaks is False

    def test_parser_accepts_short_form_output_1(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt", "-o", "output.html"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False

    def test_parser_accepts_short_form_output_2(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["-o", "output.html", "results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False

    def test_parser_accepts_long_form_output_1(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt", "--output", "output.html"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False

    def test_parser_accepts_long_form_output_2(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["--output", "output.html", "results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False

    def test_parser_takes_memory_leaks_as_a_flag(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(
            ["results.txt", "--leaks", "--output", "output.html"]
        )

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is True

    def test_parser_takes_force_flag(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(
            ["results.txt", "--force", "--output", "output.html"]
        )

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.force is True


@pytest.mark.parametrize(
    "input, expected, factory",
    (
        ("result.bin", "memray-flamegraph-result.html", FlamegraphCommand),
        ("/tmp/result.bin", "/tmp/memray-flamegraph-result.html", FlamegraphCommand),
        ("../result.bin", "../memray-flamegraph-result.html", FlamegraphCommand),
        (
            "memray-json.tool.0.bin",
            "memray-flamegraph-json.tool.0.html",
            FlamegraphCommand,
        ),
        (
            "/tmp/memray-json.tool.0.bin",
            "/tmp/memray-flamegraph-json.tool.0.html",
            FlamegraphCommand,
        ),
        (
            "../memray-json.tool.0.bin",
            "../memray-flamegraph-json.tool.0.html",
            FlamegraphCommand,
        ),
        ("memray-json.tool.0.bin", "memray-table-json.tool.0.html", TableCommand),
        ("my-result.bin", "memray-table-my-result.html", TableCommand),
        ("../my-result.bin", "../memray-table-my-result.html", TableCommand),
    ),
)
def test_determine_output(input, expected, factory):
    # GIVEN
    command = factory()

    # WHEN/THEN
    assert command.determine_output_filename(Path(input)) == Path(expected)


class TestTreeSubCommand:
    @staticmethod
    def get_prepared_parser():
        parser = argparse.ArgumentParser()
        command = TreeCommand()
        command.prepare_parser(parser)

        return command, parser

    def test_parser_rejects_no_arguments(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args([])

    def test_parser_rejects_when_no_results_provided(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args(["--biggest_allocs", "5"])

    def test_parser_accepts_single_argument(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.biggest_allocs == 200

    def test_parser_acceps_biggest_allocs_short_form(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt", "-b", "5"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.biggest_allocs == 5

    def test_parser_acceps_biggest_allocs_long_form(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt", "--biggest-allocs", "5"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.biggest_allocs == 5


class TestTableSubCommand:
    @staticmethod
    def get_prepared_parser():
        parser = argparse.ArgumentParser()
        command = TableCommand()
        command.prepare_parser(parser)

        return command, parser

    def test_parser_rejects_no_arguments(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args([])

    def test_parser_rejects_when_no_results_provided(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args(["--leaks"])

    def test_parser_accepts_single_argument(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.show_memory_leaks is False

    def test_parser_accepts_short_form_output_1(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt", "-o", "output.html"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False

    def test_parser_accepts_short_form_output_2(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["-o", "output.html", "results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False

    def test_parser_accepts_long_form_output_1(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt", "--output", "output.html"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False

    def test_parser_accepts_long_form_output_2(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["--output", "output.html", "results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False

    def test_parser_takes_memory_leaks_as_a_flag(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(
            ["results.txt", "--leaks", "--output", "output.html"]
        )

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is True

    def test_parser_takes_force_flag(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(
            ["results.txt", "--force", "--output", "output.html"]
        )

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.force is True


class TestSummarySubCommand:
    @staticmethod
    def get_prepared_parser():
        parser = argparse.ArgumentParser()
        command = SummaryCommand()
        command.prepare_parser(parser)

        return command, parser

    def test_parser_rejects_no_arguments(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args([])

    def test_parser_accepts_single_argument(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.sort_column == 1
        assert namespace.max_rows is None

    def test_parser_accepts_sort_column_long_form(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt", "--sort-column", "2"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.sort_column == 2
        assert namespace.max_rows is None

    def test_parser_accepts_sort_column_sort_form(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt", "-s", "2"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.sort_column == 2
        assert namespace.max_rows is None

    @pytest.mark.parametrize("column", [0, 12])
    def test_parser_rejects_sort_column_incorrect_values(self, column):
        # GIVEN
        command, parser = self.get_prepared_parser()

        # WHEN
        args = parser.parse_args(["results.txt", "--sort-column", str(column)])
        # THEN
        with pytest.raises(SystemExit):
            command.run(args, parser)

    def test_parser_accepts_max_rows_long_form(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt", "--max-rows", "2"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.sort_column == 1
        assert namespace.max_rows == 2

    def test_parser_accepts_max_rows_sort_form(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt", "-r", "2"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.sort_column == 1
        assert namespace.max_rows == 2


class TestStatsSubCommand:
    @staticmethod
    def get_prepared_parser():
        parser = argparse.ArgumentParser()
        command = StatsCommand()
        command.prepare_parser(parser)

        return command, parser

    def test_parser_rejects_no_arguments(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args([])

    def test_parser_accepts_single_argument(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.num_largest == 5

    def test_parser_accepts_valid_num_largest_allocators(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["-n", "3", "results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.num_largest == 3

    def test_parser_rejects_invalid_num_largest_allocators(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args(["-n", "-1", "results.txt"])


class TestTransformSubCommand:
    @staticmethod
    def get_prepared_parser():
        parser = argparse.ArgumentParser()
        command = TransformCommand()
        command.prepare_parser(parser)

        return command, parser

    def test_parser_rejects_no_arguments(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args([])

    def test_parser_rejects_when_no_results_provided(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args(["gprof2dot", "--leaks"])

    def test_parser_invalid_format(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN / THEN
        with pytest.raises(SystemExit):
            parser.parse_args(["blech"])

    def test_parser_accepts_single_argument_with_format(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["gprof2dot", "results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.format == "gprof2dot"
        assert namespace.show_memory_leaks is False

    def test_parser_accepts_short_form_output_1(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["gprof2dot", "results.txt", "-o", "output.html"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False
        assert namespace.format == "gprof2dot"

    def test_parser_accepts_short_form_output_2(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(["gprof2dot", "-o", "output.html", "results.txt"])

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False
        assert namespace.format == "gprof2dot"

    def test_parser_accepts_long_form_output_1(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(
            ["gprof2dot", "results.txt", "--output", "output.html"]
        )

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False
        assert namespace.format == "gprof2dot"

    def test_parser_accepts_long_form_output_2(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(
            ["gprof2dot", "--output", "output.html", "results.txt"]
        )

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is False
        assert namespace.format == "gprof2dot"

    def test_parser_takes_memory_leaks_as_a_flag(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(
            ["gprof2dot", "results.txt", "--leaks", "--output", "output.html"]
        )

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.show_memory_leaks is True
        assert namespace.format == "gprof2dot"

    def test_parser_takes_force_flag(self):
        # GIVEN
        _, parser = self.get_prepared_parser()

        # WHEN
        namespace = parser.parse_args(
            ["gprof2dot", "results.txt", "--force", "--output", "output.html"]
        )

        # THEN
        assert namespace.results == "results.txt"
        assert namespace.output == "output.html"
        assert namespace.force is True
        assert namespace.format == "gprof2dot"


class TestRunCommandReal:
    def test_run_non_existent_file(self, capsys):
        # GIVEN
        # WHEN
        exit_code = main(["run", "non_existent_file.py"])

        # THEN
        assert exit_code == 1
        captured = capsys.readouterr()
        assert "No such file or directory: 'non_existent_file.py'" in captured.err

    def test_run_non_existent_module(self, capsys):
        # GIVEN
        # WHEN
        exit_code = main(["run", "-m", "non_existent_module"])

        # THEN
        assert exit_code == 1
        captured = capsys.readouterr()
        assert "No module named non_existent_module" in captured.err
        assert "[memray] Successfully generated profile results." not in captured.out



import sys


def fib1(n):
    my_list = [0, 1]
    for i in range(2, n + 1):
        my_list.append(my_list[i - 1] + my_list[i - 2])
    return my_list[-1]


def fib2(n, cache={0: 0, 1: 1}):
    if n in cache:
        return cache[n]
    cache[n] = fib2(n - 1) + fib2(n - 2)
    return cache[n]


def run():
    sys.setrecursionlimit(100000)
    n = 99900
    a = fib1(n)
    b = fib2(n)

    assert a == b


run()



"""Tests for our testing utilities."""

from memray import AllocationRecord
from memray import AllocatorType
from tests.utils import MockAllocationRecord
from tests.utils import filter_relevant_allocations


class TestFilterRelevantAllocations:
    def test_filters_for_valloc_and_free(self):
        records = [
            MockAllocationRecord(1, 0x1000000, 1024, AllocatorType.MALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 1024, AllocatorType.VALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, []),
        ]
        assert list(filter_relevant_allocations(records)) == [
            MockAllocationRecord(1, 0x1000000, 1024, AllocatorType.VALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, []),
        ]

    def test_filters_based_on_addresses(self):
        records = [
            MockAllocationRecord(1, 0x2000000, 1024, AllocatorType.MALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 1024, AllocatorType.VALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x2000000, 0, AllocatorType.FREE, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, []),
        ]
        assert list(filter_relevant_allocations(records)) == [
            MockAllocationRecord(1, 0x1000000, 1024, AllocatorType.VALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, []),
        ]

    def test_free_records_with_valid_addresses_that_dont_match_do_not_appear(self):
        records = [
            MockAllocationRecord(1, 0x2000000, 0, AllocatorType.FREE, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, []),
            MockAllocationRecord(1, 0x2000000, 1024, AllocatorType.MALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 1024, AllocatorType.VALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x2000000, 0, AllocatorType.FREE, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, []),
        ]
        assert list(filter_relevant_allocations(records)) == [
            MockAllocationRecord(1, 0x1000000, 1024, AllocatorType.VALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, []),
        ]

    def test_free_records_with_unmatched_addresses_do_not_appear(self):
        records = [
            MockAllocationRecord(1, 0x6000000, 0, AllocatorType.FREE, 0, 0, []),
            MockAllocationRecord(1, 0x7000000, 0, AllocatorType.FREE, 0, 0, []),
            MockAllocationRecord(1, 0x2000000, 1024, AllocatorType.MALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 1024, AllocatorType.VALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x2000000, 0, AllocatorType.FREE, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, []),
        ]
        assert list(filter_relevant_allocations(records)) == [
            MockAllocationRecord(1, 0x1000000, 1024, AllocatorType.VALLOC, 0, 0, []),
            MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, []),
        ]


class TestMockAllocationRecord:
    def test_holds_values_at_correct_names(self):
        tid = 0
        address = 0x1140000
        size = 1024
        allocator = AllocatorType.MALLOC
        stack_id = 1
        n_allocations = 1
        stack = ["stack 0", "stack 1"]

        mock_record = MockAllocationRecord(
            tid, address, size, allocator, stack_id, n_allocations, stack
        )
        assert mock_record.tid == tid
        assert mock_record.address == address
        assert mock_record.size == size
        assert mock_record.allocator == allocator
        assert mock_record.stack_id == stack_id
        assert mock_record.n_allocations == n_allocations
        assert mock_record.stack_trace() == stack
        assert mock_record.stack_trace(max_stacks=1) == ["stack 0"]

    def test_looks_like_AllocationRecord(self):
        tid = 0
        address = 0x1140000
        size = 1024
        allocator = AllocatorType.MALLOC
        stack_id = 1
        n_allocations = 1
        stack = ["i'm stack"]

        record = AllocationRecord(
            (tid, address, size, allocator, stack_id, n_allocations)
        )
        mock_record = MockAllocationRecord(
            tid, address, size, allocator, stack_id, n_allocations, stack
        )

        assert mock_record.tid == record.tid == tid
        assert mock_record.address == record.address == address
        assert mock_record.size == record.size == size
        assert mock_record.allocator == record.allocator == allocator
        assert mock_record.stack_id == record.stack_id == stack_id
        assert mock_record.n_allocations == record.n_allocations == n_allocations

    def test_equality(self):
        one = MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, [])
        two = MockAllocationRecord(1, 0x1000000, 0, AllocatorType.FREE, 0, 0, [])
        three = MockAllocationRecord(2, 0x1000000, 0, AllocatorType.FREE, 0, 0, [])

        assert one == one
        assert one == two
        assert one != three

        assert two == one
        assert two == two
        assert two != three

        assert three != one
        assert three != two
        assert three == three



import dataclasses
import json
import pathlib
import subprocess
import sys
import tempfile
from typing import List

from .plot import plot_diff

CASES_DIR = pathlib.Path(__file__).parent / "cases"
RESULTS_DIR = pathlib.Path(__file__).parent / "results"
DOCUTILS_DATA = pathlib.Path(__file__).parent / "cases" / "docutils_data" / "docs"
TELCO_DATA = pathlib.Path(__file__).parent / "cases" / "telco_data" / "telco-bench.b"


@dataclasses.dataclass
class Case:
    name: str
    file_name: str
    arguments: List[str] = dataclasses.field(default_factory=list)

    def run_case(
        self,
        run_name: str,
        template_file: pathlib.Path,
        tracker_options: str,
        results_file: str,
    ) -> None:
        with tempfile.TemporaryDirectory() as tmpdirname:
            case_file = pathlib.Path(tmpdirname) / f"{self.file_name}.py"
            code = template_file.read_text()
            case_file.write_text(code)

            helper_file = pathlib.Path(tmpdirname) / "memray_helper.py"
            helper_code = (
                "import memray\nimport contextlib\n"
                f"def get_tracker():\n    return {tracker_options}"
            )
            helper_file.write_text(helper_code)

            results_file = RESULTS_DIR / results_file
            print(f"Running {self.name} with arguments {self.arguments} - {run_name}")
            subprocess.run(
                [sys.executable, case_file, "-o", results_file] + self.arguments,
                check=True,
            )

    def run(self) -> None:
        template_file = CASES_DIR / f"{self.file_name}_memray.py"
        if not template_file.exists():
            raise ValueError(f"Case {self.name} does not exist.")

        self.run_case(
            "", template_file, "contextlib.nullcontext()", f"{self.name}.json"
        )

        self.run_case(
            "memray base",
            template_file,
            "memray.Tracker('/dev/null')",
            f"{self.name}_memray.json",
        )

        self.run_case(
            "memray python allocators",
            template_file,
            "memray.Tracker('/dev/null', trace_python_allocators=True)",
            f"{self.name}_memray_python_allocators.json",
        )

        self.run_case(
            "memray python native",
            template_file,
            "memray.Tracker('/dev/null', native_traces=True)",
            f"{self.name}_memray_python_native.json",
        )

        self.run_case(
            "memray python all",
            template_file,
            "memray.Tracker('/dev/null', trace_python_allocators=True, native_traces=True)",
            f"{self.name}_memray_python_all.json",
        )


CASES = [
    Case("docutils", "docutils_html", [f"--doc_root={DOCUTILS_DATA}"]),
    Case("raytrace", "raytrace", []),
    Case("fannkuch", "fannkuch", []),
    Case("pprint", "pprint_format", []),
    Case("mdp", "mdp", []),
    Case("async_tree", "async_tree", ["none"]),
    Case("async_tree_io", "async_tree", ["io"]),
    Case("async_tree_mem", "async_tree", ["memoization"]),
    Case("async_tree_cpu_io", "async_tree", ["cpu_io_mixed"]),
    Case("deltablue", "deltablue", []),
    Case("nbody", "nbody", []),
    Case("nqueens", "nqueens", []),
    Case("regex_dna", "regex_dna", []),
    Case("go", "go", []),
    Case("hexion", "hexion", []),
    Case("meteor_context", "meteor_context", []),
    Case("json_dumps", "json_dumps", []),
    Case("json_loads", "json_loads", []),
    Case(
        "picke_pure_python",
        "pickles",
        [
            "pickle",
            "--pure-python",
        ],
    ),
    Case("picke", "pickles", ["pickle"]),
    Case(
        "unpicke_pure_python",
        "pickles",
        [
            "unpickle",
            "--pure-python",
        ],
    ),
    Case("unpicke", "pickles", ["unpickle"]),
    Case(
        "pickle_list_pure_python",
        "pickles",
        [
            "pickle_list",
            "--pure-python",
        ],
    ),
    Case("pickle_list", "pickles", ["pickle_list"]),
    Case(
        "unpickle_list_pure_python",
        "pickles",
        [
            "unpickle_list",
            "--pure-python",
        ],
    ),
    Case("unpickle_list", "pickles", ["unpickle_list"]),
    Case(
        "pickle_dict_pure_python",
        "pickles",
        [
            "pickle_dict",
            "--pure-python",
        ],
    ),
    Case("pickle_dict", "pickles", ["pickle_dict"]),
    Case("spectral_norm", "spectral_norm", []),
    Case("telco", "telco", [f"--doc_root={TELCO_DATA}"]),
    Case("sqlite_synth", "sqlite_synth", []),
    Case("regex_v8", "regex_v8", []),
    Case("regex_effbot", "regex_effbot", []),
    Case("regex_effbot_bytes", "regex_effbot", ["--force_bytes"]),
]


@dataclasses.dataclass
class BenchmarkResult:
    name: str
    data: List


def gather_benchmarks(cases):
    results = []
    names = ("", "Defaut", "Python allocators", "Native", "Python allocators + Native")
    extensions = (
        "",
        "_memray",
        "_memray_python_allocators",
        "_memray_python_native",
        "_memray_python_all",
    )
    for name, extension in zip(names, extensions):
        type_results = []
        for case in cases:
            memray_results_file = RESULTS_DIR / f"{case.name}{extension}.json"
            data = json.loads(memray_results_file.read_text())
            type_results.append(data)
        results.append(BenchmarkResult(name=name, data=type_results))
    return results


if __name__ == "__main__":
    # if RESULTS_DIR.exists():
    #     raise RuntimeError(f"Results directory {RESULTS_DIR} already exists")

    # RESULTS_DIR.mkdir()

    for case in CASES:
        case.run()

    base_results, *memray_results = gather_benchmarks(CASES)
    for memray_result in memray_results:
        plot_diff(
            memray_result,
            base_results,
            f"plot_{memray_result.name}.png".replace(" ", "_").lower(),
            f"Overhead of Memray with {memray_result.name}",
        )



import pytest


@pytest.fixture(autouse=True)
def use_80_columns(monkeypatch):
    """Override the COLUMNS environment variable to 80.

    This matches the assumed terminal width that is hardcoded in the tests.
    """
    monkeypatch.setenv("COLUMNS", "80")



from dataclasses import dataclass

from memray import AllocatorType
from memray._memray import AllocationLifetimeAggregatorTestHarness
from memray._memray import Interval

CALLOC = AllocatorType.CALLOC
FREE = AllocatorType.FREE
MMAP = AllocatorType.MMAP
MUNMAP = AllocatorType.MUNMAP


@dataclass(frozen=True)
class Location:
    tid: int
    native_frame_id: int
    frame_index: int
    native_segment_generation: int


def test_no_allocations_at_start():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()

    # WHEN
    # THEN
    assert [] == list(tester.get_allocations())


def test_allocation_not_reported_when_freed_within_same_snapshot():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)

    # THEN
    assert [] == list(tester.get_allocations())


def test_allocation_reported_when_freed_within_different_snapshot():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, 1, 1, 1234)]


def test_allocation_reported_when_leaked():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, None, 1, 1234)]


def test_multiple_snapshots_between_allocation_and_deallocation():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.capture_snapshot()
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.capture_snapshot()
    tester.capture_snapshot()
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)
    tester.capture_snapshot()

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(2, 5, 1, 1234)]


def test_allocations_from_same_location_and_snapshot_freed_in_different_snapshots():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=4321)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(1, 2, 1, 4321), Interval(1, 3, 1, 1234)]


def test_allocations_from_same_location_and_different_snapshots_freed_in_one_snapshot():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=4321)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, 2, 1, 1234), Interval(1, 2, 1, 4321)]


def test_two_leaked_allocations_from_one_location():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=4321)
    tester.capture_snapshot()

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, None, 1, 1234), Interval(1, None, 1, 4321)]


def test_allocations_made_and_freed_together_are_aggregated():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=4321)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, 1, 2, 1234 + 4321)]


def test_leaked_allocations_within_one_snapshot_are_aggregated():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=4321)
    tester.capture_snapshot()

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, None, 2, 1234 + 4321)]


def test_freed_allocations_from_different_locations_are_not_aggregated():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=1,
        native_frame_id=7,
        frame_index=8,
        native_segment_generation=9,
    )
    free = Location(
        tid=0,
        native_frame_id=0,
        frame_index=0,
        native_segment_generation=0,
    )

    # WHEN
    tester.add_allocation(**loc1.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc2.__dict__, allocator=CALLOC, address=8192, size=4321)
    tester.capture_snapshot()
    tester.add_allocation(**free.__dict__, allocator=FREE, address=8192, size=0)
    tester.add_allocation(**free.__dict__, allocator=FREE, address=4096, size=0)

    # THEN
    alloc1, alloc2 = tester.get_allocations()
    assert alloc1.allocator == CALLOC
    assert alloc1.native_stack_id == 4
    assert alloc1.stack_id == 5
    assert alloc1.native_segment_generation == 6
    assert alloc1.tid == 1
    assert alloc1.intervals == [Interval(0, 1, 1, 1234)]

    assert alloc2.allocator == CALLOC
    assert alloc2.native_stack_id == 7
    assert alloc2.stack_id == 8
    assert alloc2.native_segment_generation == 9
    assert alloc2.tid == 1
    assert alloc2.intervals == [Interval(0, 1, 1, 4321)]


def test_leaked_allocations_from_different_locations_are_not_aggregated():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=1,
        native_frame_id=7,
        frame_index=8,
        native_segment_generation=9,
    )

    # WHEN
    tester.add_allocation(**loc1.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc2.__dict__, allocator=CALLOC, address=8192, size=4321)

    # THEN
    alloc1, alloc2 = tester.get_allocations()
    assert alloc1.allocator == CALLOC
    assert alloc1.native_stack_id == 4
    assert alloc1.stack_id == 5
    assert alloc1.native_segment_generation == 6
    assert alloc1.tid == 1
    assert alloc1.intervals == [Interval(0, None, 1, 1234)]

    assert alloc2.allocator == CALLOC
    assert alloc2.native_stack_id == 7
    assert alloc2.stack_id == 8
    assert alloc2.native_segment_generation == 9
    assert alloc2.tid == 1
    assert alloc2.intervals == [Interval(0, None, 1, 4321)]


def test_range_freed_in_same_snapshot():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=MMAP, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=MUNMAP, address=4096, size=1234)

    # THEN
    assert [] == list(tester.get_allocations())


def test_range_freed_in_different_snapshot():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=MMAP, address=4096, size=1234)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=MUNMAP, address=4096, size=1234)

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == MMAP
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, 1, 1, 1234)]


def test_range_leaked():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=MMAP, address=4096, size=1234)

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == MMAP
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, None, 1, 1234)]


def test_shrunk_then_leaked_range():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=MMAP, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=MUNMAP, address=4096, size=1000)

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == MMAP
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, None, 1, 234)]


def test_shrunk_then_freed_range():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=MMAP, address=4096, size=1234)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=MUNMAP, address=4096, size=1000)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=MUNMAP, address=4096, size=1234)

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == MMAP
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, 1, 0, 1000), Interval(0, 2, 1, 234)]


def test_split_then_leaked_range():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=MMAP, address=4096, size=1234)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=MUNMAP, address=5000, size=100)

    # THEN
    (alloc,) = tester.get_allocations()

    assert alloc.allocator == MMAP
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [Interval(0, 1, 0, 100), Interval(0, None, 1, 1234 - 100)]


def test_split_then_freed_range():
    # GIVEN
    tester = AllocationLifetimeAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=MMAP, address=4096, size=1234)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=MUNMAP, address=5000, size=100)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=MUNMAP, address=4096, size=904)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=MUNMAP, address=5100, size=230)

    # THEN
    (alloc,) = tester.get_allocations()
    assert alloc.allocator == MMAP
    assert alloc.native_stack_id == 4
    assert alloc.stack_id == 5
    assert alloc.native_segment_generation == 6
    assert alloc.tid == 1
    assert alloc.intervals == [
        Interval(0, 1, 0, 100),
        Interval(0, 2, 0, 904),
        Interval(0, 3, 1, 230),
    ]



import numpy as np
import pytest
from exercise_2.holding_onto_memory import process_data


@pytest.mark.limit_memory("230 MB")
def test_holding_in_memory():
    process_data()


def test_result():
    result = process_data()
    assert np.all(result == 18)



import mmap
import os
import tempfile

from memray import AllocatorType
from memray import FileReader

try:
    from memray._test import MemoryAllocator
except ImportError:
    from memray import MemoryAllocator

from memray import Tracker

from .benchmarking.cases import async_tree_base
from .benchmarking.cases import fannkuch_base
from .benchmarking.cases import mdp_base
from .benchmarking.cases import pprint_format_base
from .benchmarking.cases import raytrace_base

LOOPS = 1000


class TracebackBenchmarks:
    def setup(self):
        self.tempfile = tempfile.NamedTemporaryFile()
        allocator = MemoryAllocator()

        def fac(n):
            if n == 1:
                allocator.valloc(1234)
                allocator.free()
                return 1
            return n * fac(n - 1)

        os.unlink(self.tempfile.name)
        with Tracker(self.tempfile.name):
            fac(300)

        (self.record,) = [
            record
            for record in FileReader(self.tempfile.name).get_allocation_records()
            if record.allocator == AllocatorType.VALLOC
        ]

    def time_get_stack_trace(self):
        self.record.stack_trace()


class AllocatorBenchmarks:
    def setup(self):
        self.tempfile = tempfile.NamedTemporaryFile()
        self.allocator = MemoryAllocator()

    def time_malloc(self):
        os.unlink(self.tempfile.name)
        with Tracker(self.tempfile.name):
            for _ in range(LOOPS):
                if self.allocator.malloc(1234):
                    self.allocator.free()

    def time_posix_memalign(self):
        os.unlink(self.tempfile.name)
        with Tracker(self.tempfile.name):
            for _ in range(LOOPS):
                if self.allocator.posix_memalign(1234):
                    self.allocator.free()

    def time_posix_realloc(self):
        os.unlink(self.tempfile.name)
        with Tracker(self.tempfile.name):
            for _ in range(LOOPS):
                if self.allocator.posix_memalign(1234):
                    self.allocator.free()

    def time_calloc(self):
        os.unlink(self.tempfile.name)
        with Tracker(self.tempfile.name):
            for _ in range(LOOPS):
                if self.allocator.calloc(1234):
                    self.allocator.free()

    def time_pvalloc(self):
        os.unlink(self.tempfile.name)
        with Tracker(self.tempfile.name):
            for _ in range(LOOPS):
                if self.allocator.pvalloc(1234):
                    self.allocator.free()

    def time_valloc(self):
        os.unlink(self.tempfile.name)
        with Tracker(self.tempfile.name):
            for _ in range(LOOPS):
                if self.allocator.valloc(1234):
                    self.allocator.free()

    def time_realloc(self):
        os.unlink(self.tempfile.name)
        with Tracker(self.tempfile.name):
            for _ in range(LOOPS):
                if self.allocator.realloc(1234):
                    self.allocator.free()

    def time_mmap(self):
        os.unlink(self.tempfile.name)
        with Tracker(self.tempfile.name):
            for _ in range(LOOPS):
                with mmap.mmap(-1, length=2048, access=mmap.ACCESS_WRITE) as mmap_obj:
                    mmap_obj[0:100] = b"a" * 100


class ParserBenchmarks:
    def setup(self):
        self.tempfile = tempfile.NamedTemporaryFile()
        self.allocator = MemoryAllocator()
        os.unlink(self.tempfile.name)
        self.tracker = Tracker(self.tempfile.name)
        with self.tracker:
            for _ in range(LOOPS):
                self.allocator.valloc(1234)
                self.allocator.free()

    def time_end_to_end_parsing(self):
        list(FileReader(self.tempfile.name).get_allocation_records())


def recursive(n, chunk_size):
    """Mimics generally-increasing but spiky usage"""
    if not n:
        return

    allocator = MemoryAllocator()
    allocator.valloc(n * chunk_size)

    # Don't keep allocated memory when recursing, ~50% of the calls.
    if n % 2:
        allocator.free()
        recursive(n - 1, chunk_size)
    else:
        recursive(n - 1, chunk_size)
        allocator.free()


class HighWatermarkBenchmarks:
    def setup(self):
        self.tempfile = tempfile.NamedTemporaryFile()
        os.unlink(self.tempfile.name)
        self.tracker = Tracker(self.tempfile.name)

        with self.tracker:
            recursive(700, 99)

    def time_high_watermark(self):
        list(
            FileReader(self.tempfile.name).get_high_watermark_allocation_records(
                merge_threads=False
            )
        )


class MacroBenchmarksBase:
    def __init_subclass__(cls) -> None:
        for name in dir(cls):
            if name.startswith("bench_"):
                bench_name = name[len("bench_") :]
                setattr(cls, f"time_{bench_name}", getattr(cls, name))

    def bench_async_tree_cpu(self):
        with self.tracker:
            async_tree_base.run_benchmark("none")

    def bench_async_tree_io(self):
        with self.tracker:
            async_tree_base.run_benchmark("io")

    def bench_async_tree_memoization(self):
        with self.tracker:
            async_tree_base.run_benchmark("memoization")

    def bench_async_tree_cpu_io_mixed(self):
        with self.tracker:
            async_tree_base.run_benchmark("cpu_io_mixed")

    def bench_fannkuch(self):
        with self.tracker:
            fannkuch_base.run_benchmark()

    def bench_mdp(self):
        with self.tracker:
            mdp_base.run_benchmark()

    def bench_pprint_format(self):
        with self.tracker:
            pprint_format_base.run_benchmark()

    def bench_raytrace(self):
        with self.tracker:
            raytrace_base.run_benchmark()


class MacroBenchmarksDefault(MacroBenchmarksBase):
    def setup(self):
        self.tracker_args = ("/dev/null",)
        self.tracker_kwargs = {}
        self.tracker = Tracker(*self.tracker_args, **self.tracker_kwargs)


class MacroBenchmarksPythonAllocators(MacroBenchmarksBase):
    def setup(self):
        self.tracker_args = ("/dev/null",)
        self.tracker_kwargs = {"trace_python_allocators": True}
        self.tracker = Tracker(*self.tracker_args, **self.tracker_kwargs)


class MacroBenchmarksPythonNative(MacroBenchmarksBase):
    def setup(self):
        self.tracker_args = ("/dev/null",)
        self.tracker_kwargs = {"native_traces": True}
        self.tracker = Tracker(*self.tracker_args, **self.tracker_kwargs)


class MacroBenchmarksPythonAll(MacroBenchmarksBase):
    def setup(self):
        self.tracker_args = ("/dev/null",)
        self.tracker_kwargs = {"native_traces": True, "trace_python_allocators": True}
        self.tracker = Tracker(*self.tracker_args, **self.tracker_kwargs)


class FileSizeBenchmarks:
    def setup(self):
        self.tempfile = tempfile.NamedTemporaryFile()
        os.unlink(self.tempfile.name)
        self.tracker = Tracker(self.tempfile.name)

    def track_async_tree_cpu(self):
        with self.tracker:
            async_tree_base.run_benchmark("none")
        return os.stat(self.tempfile.name).st_size

    def track_async_tree_io(self):
        with self.tracker:
            async_tree_base.run_benchmark("io")
        return os.stat(self.tempfile.name).st_size

    def track_async_tree_memoization(self):
        with self.tracker:
            async_tree_base.run_benchmark("memoization")
        return os.stat(self.tempfile.name).st_size

    def track_async_tree_cpu_io_mixed(self):
        with self.tracker:
            async_tree_base.run_benchmark("cpu_io_mixed")
        return os.stat(self.tempfile.name).st_size

    def track_fannkuch(self):
        with self.tracker:
            fannkuch_base.run_benchmark()
        return os.stat(self.tempfile.name).st_size

    def track_mdp(self):
        with self.tracker:
            mdp_base.run_benchmark()
        return os.stat(self.tempfile.name).st_size

    def track_pprint_format(self):
        with self.tracker:
            pprint_format_base.run_benchmark()
        return os.stat(self.tempfile.name).st_size

    def track_raytrace(self):
        with self.tracker:
            raytrace_base.run_benchmark()
        return os.stat(self.tempfile.name).st_size



import asyncio
import datetime
from io import StringIO
from typing import Awaitable
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import cast

import pytest
from rich import print as rprint
from textual.app import App
from textual.coordinate import Coordinate
from textual.pilot import Pilot
from textual.widget import Widget
from textual.widgets import DataTable
from textual.widgets import Label

import memray.reporters.tui
from memray import AllocationRecord
from memray import AllocatorType
from memray.reporters.tui import Location
from memray.reporters.tui import MemoryGraph
from memray.reporters.tui import Snapshot
from memray.reporters.tui import SnapshotFetched
from memray.reporters.tui import TUIApp
from memray.reporters.tui import aggregate_allocations
from tests.utils import MockAllocationRecord
from tests.utils import async_run


class MockApp(TUIApp):
    CSS_PATH = None  # type: ignore

    def __init__(self, *args, disable_update_thread=True, **kwargs):
        super().__init__(*args, **kwargs)
        if disable_update_thread:
            # Make the update thread return immediately when started
            self._update_thread.cancel()

    def add_mock_snapshot(
        self,
        snapshot: List[MockAllocationRecord],
        disconnected: bool = False,
        native: bool = True,
    ) -> None:
        records = cast(List[AllocationRecord], snapshot)
        self.post_message(
            SnapshotFetched(
                Snapshot(
                    heap_size=sum(record.size for record in records),
                    records=records,
                    records_by_location=aggregate_allocations(
                        cast(List[AllocationRecord], records), native_traces=native
                    ),
                ),
                disconnected,
            )
        )

    def add_mock_snapshots(
        self,
        snapshots: List[List[MockAllocationRecord]],
        disconnect_after_last: bool = True,
        native: bool = True,
    ) -> None:
        for i, snapshot in enumerate(snapshots):
            disconnected = i == len(snapshots) - 1 and disconnect_after_last
            self.add_mock_snapshot(snapshot, disconnected=disconnected, native=native)


class MockReader:
    def __init__(
        self,
        snapshots: List[List[MockAllocationRecord]],
        has_native_traces: bool = True,
        pid: Optional[int] = None,
        command_line: Optional[str] = None,
    ):
        self._snapshots = cast(List[List[AllocationRecord]], snapshots)
        self._next_snapshot = 0
        self.is_active = True
        self.command_line = command_line
        self.pid = pid
        self.has_native_traces = has_native_traces

    def get_current_snapshot(
        self, *, merge_threads: bool
    ) -> Iterable[AllocationRecord]:
        assert isinstance(merge_threads, bool)  # ignore unused argument
        assert self.is_active
        snapshot = self._snapshots[self._next_snapshot]
        self._next_snapshot += 1
        self.is_active = self._next_snapshot < len(self._snapshots)
        return snapshot


@pytest.fixture
def compare(monkeypatch, tmp_path, snap_compare):
    monkeypatch.setattr(memray.reporters.tui, "datetime", FakeDatetime)

    def compare_impl(
        cmdline_override: Optional[str] = None,
        press: Iterable[str] = (),
        terminal_size: Tuple[int, int] = (80, 24),
        run_before: Optional[Callable[[Pilot], Optional[Awaitable[None]]]] = None,
        native: bool = True,
    ):
        async def run_before_wrapper(pilot) -> None:
            if run_before is not None:
                result = run_before(pilot)
                if result is not None:
                    await result

            await pilot.pause()
            await pilot.pause(0.2)
            header = pilot.app.screen.query_one("Header")
            header.last_update = header.start + datetime.timedelta(seconds=42)

        app = MockApp(
            MockReader([], has_native_traces=native),
            cmdline_override=cmdline_override,
        )

        try:
            asyncio.get_event_loop()
        except RuntimeError:
            asyncio.set_event_loop(asyncio.new_event_loop())

        return snap_compare(
            app,
            press=press,
            terminal_size=terminal_size,
            run_before=run_before_wrapper,
        )

    yield compare_impl


def render_widget(widget: Widget) -> str:
    output = StringIO()
    content = getattr(widget, "renderable", None)
    if content is None:
        content = widget.content  # type: ignore
    assert content is not None
    rprint(content, file=output)
    return output.getvalue()


def extract_label_text(app: App) -> Dict[str, str]:
    return {
        label.id: render_widget(label)
        for label in app.screen.query(Label)
        if label.id is not None
    }


def mock_allocation(
    stack: Optional[List[Tuple[str, str, int]]] = None,
    tid: int = 1,
    address: int = 0,
    size: int = 1024,
    allocator: AllocatorType = AllocatorType.MALLOC,
    stack_id: int = 0,
    n_allocations: int = 1,
    thread_name: str = "",
):
    hybrid_stack = stack

    if hybrid_stack is not None:
        stack = [
            (func, filename, lineno)
            for func, filename, lineno in hybrid_stack
            if filename.endswith(".py")
        ]

    return MockAllocationRecord(
        tid=tid,
        address=address,
        size=size,
        allocator=allocator,
        stack_id=stack_id,
        n_allocations=n_allocations,
        thread_name=thread_name,
        _stack=stack,
        _hybrid_stack=hybrid_stack,
    )


SHORT_SNAPSHOTS = [
    [
        mock_allocation(
            stack=[
                ("malloc", "malloc.c", 1234),
                ("f1", "f.py", 16),
                ("parent", "fun.py", 8),
                ("grandparent", "fun.py", 4),
            ],
        ),
    ],
    [
        mock_allocation(
            stack=[
                ("malloc", "malloc.c", 1234),
                ("f1", "f.py", 16),
                ("parent", "fun.py", 8),
                ("grandparent", "fun.py", 4),
            ],
        ),
        mock_allocation(
            stack=[
                ("malloc", "malloc.c", 1234),
                ("f2", "f.py", 32),
                ("parent", "fun.py", 8),
                ("grandparent", "fun.py", 4),
            ],
        ),
        mock_allocation(
            stack=[
                ("malloc", "malloc.c", 1234),
                ("f2", "f.py", 32),
                ("parent", "fun.py", 8),
                ("grandparent", "fun.py", 4),
            ],
        ),
    ],
]


LONG_SNAPSHOTS = [
    [
        mock_allocation(
            stack=[
                ("malloc", "malloc.c", 1234),
                ("f1", "f.py", 16),
                ("parent", "fun.py", 8),
                ("grandparent", "fun.py", 4),
            ],
        ),
    ],
    [
        mock_allocation(
            stack=[
                ("malloc", "malloc.c", 1234),
                ("f1", "f.py", 16),
                ("parent", "fun.py", 8),
                ("grandparent", "fun.py", 4),
            ],
        ),
        mock_allocation(
            stack=[
                ("malloc", "malloc.c", 1234),
                ("f2", "f.py", 32),
                ("parent", "fun.py", 8),
                ("grandparent", "fun.py", 4),
            ],
        ),
        mock_allocation(
            size=333,
            stack=[
                ("malloc", "malloc.c", 1234),
                *[(f"something{i}", "something.py", i) for i in range(20)],
                ("f2", "f.py", 32),
                ("parent", "fun.py", 8),
                ("grandparent", "fun.py", 4),
            ],
        ),
    ],
]


class FakeDatetime(datetime.datetime):
    @classmethod
    def now(cls):
        return cls(2023, 10, 13, 12)


class TestGraph:
    def test_empty(self):
        # GIVEN

        plot = MemoryGraph(max_data_points=50)

        # WHEN

        graph = tuple(plot.render_line(i).text for i in range(plot._height))

        # THEN

        assert plot._maxval == 1.0
        assert plot._minval == 0.0
        assert graph == (" " * 50, " " * 50, " " * 50, " " * 50)

    def test_size_of_graph(self):
        # GIVEN
        size = 36
        rows = 10

        plot = MemoryGraph(max_data_points=size, height=rows)

        # WHEN

        for point in range(50):
            plot.add_value(point)
        graph = tuple(plot.render_line(i).text for i in range(plot._height))

        # THEN

        assert len(graph) == rows
        assert all(len(line) == size for line in graph)

    def test_one_point_lower_than_max(self):
        # GIVEN

        plot = MemoryGraph(max_data_points=50)

        # WHEN

        plot.add_value(0.5)
        graph = tuple(plot.render_line(i).text for i in range(plot._height))

        # THEN

        assert plot._maxval == 1.0
        assert plot._minval == 0.0
        assert graph == (
            "                                                  ",
            "                                                  ",
            "                                                 ▐",
            "                                                 ▐",
        )

    def test_one_point_bigger_than_max(self):
        # GIVEN

        plot = MemoryGraph(max_data_points=50)

        # WHEN

        plot.add_value(500.0)
        graph = tuple(plot.render_line(i).text for i in range(plot._height))

        # THEN

        assert plot._maxval == 500.0
        assert plot._minval == 0
        assert graph == (
            "                                                 ▐",
            "                                                 ▐",
            "                                                 ▐",
            "                                                 ▐",
        )

    def test_one_point_bigger_than_max_after_resize(self):
        # GIVEN

        plot = MemoryGraph(max_data_points=50)

        # WHEN
        plot.add_value(1000)
        for _ in range(50 * 2):
            plot.add_value(0)
        plot.add_value(500.0)

        graph = tuple(plot.render_line(i).text for i in range(plot._height))

        # THEN

        assert plot._maxval == 1000.0
        assert plot._minval == 0
        assert graph == (
            "                                                  ",
            "                                                  ",
            "                                                 ▐",
            "                                                 ▐",
        )

    def test_multiple_points(self):
        # GIVEN

        plot = MemoryGraph(max_data_points=50)
        plot.add_value(100.0)
        for _ in range(50 * 2):
            plot.add_value(0)

        # WHEN

        for point in range(50):
            plot.add_value(point)

        graph = tuple(plot.render_line(i).text for i in range(plot._height))

        # THEN

        assert plot._maxval == 100.0
        assert plot._minval == 0
        assert graph == (
            "                                                  ",
            "                                                  ",
            "                                      ▄▄▄▄▄▄██████",
            "                         ▗▄▄▄▄▄▟██████████████████",
        )

    def test_multiple_points_scattered(self):
        # GIVEN

        plot = MemoryGraph(max_data_points=50)
        plot.add_value(100.0)
        for _ in range(50 * 2):
            plot.add_value(0)

        # WHEN
        plot.add_value(100)
        plot.add_value(15)
        plot.add_value(30)
        plot.add_value(75)

        graph = tuple(plot.render_line(i).text for i in range(plot._height))

        # THEN

        assert plot._maxval == 100.0
        assert plot._minval == 0

        assert graph == (
            "                                                ▌ ",
            "                                                ▌▐",
            "                                                ▌▟",
            "                                                ██",
        )


@pytest.mark.parametrize("native_traces", [False, True])
def test_update_thread(native_traces):
    """Test that our update thread posts the expected messages to our app."""
    # GIVEN
    snapshots = SHORT_SNAPSHOTS
    reader = MockReader(snapshots, native_traces)
    messages = []
    all_messages_received = asyncio.Event()

    class MessageInterceptingApp(MockApp):
        def __init__(self, reader):
            super().__init__(reader, poll_interval=0.01, disable_update_thread=False)

        def on_snapshot_fetched(self, message):
            messages.append(message)
            if message.disconnected:
                all_messages_received.set()

    app = MessageInterceptingApp(reader)

    # WHEN
    async def run_test():
        async with app.run_test() as pilot:
            await all_messages_received.wait()
            await pilot.pause()

    async_run(run_test())

    # THEN
    assert len(messages) == len(snapshots)
    for i, message in enumerate(messages):
        last_message = i == len(messages) - 1
        assert message.disconnected is last_message
        assert message.snapshot.heap_size == sum(a.size for a in snapshots[i])
        assert message.snapshot.records == snapshots[i]
        assert message.snapshot.records_by_location == aggregate_allocations(
            message.snapshot.records,
            native_traces=native_traces,
        )


@pytest.mark.parametrize(
    "pid, display_val",
    [
        pytest.param(999, "PID: 999", id="Known PID"),
        pytest.param(None, "PID: ???", id="Unknown PID"),
    ],
)
def test_pid_display(pid, display_val):
    # GIVEN
    reader = MockReader([], pid=pid)
    app = MockApp(reader)
    labels = {}

    # WHEN
    async def run_test():
        async with app.run_test() as pilot:
            await pilot.pause()
            labels.update(extract_label_text(pilot.app))

    async_run(run_test())

    # THEN
    assert labels["pid"].rstrip() == display_val


@pytest.mark.parametrize(
    "command_line, display_val",
    [
        pytest.param("foo bar baz", "CMD: foo bar baz", id="Known command"),
        pytest.param(
            "/path/to/foo bar baz",
            "CMD: /path/to/foo bar baz",
            id="Known command with path",
        ),
        pytest.param(
            "/path/to/memray bar baz",
            "CMD: memray bar baz",
            id="Memray script with path",
        ),
        pytest.param(
            "/path/to/memray/__main__.py bar baz",
            "CMD: memray bar baz",
            id="Memray module with path",
        ),
        pytest.param(None, "CMD: ???", id="Unknown command"),
    ],
)
def test_command_line_display(command_line, display_val):
    # GIVEN
    reader = MockReader([], command_line=command_line)
    app = MockApp(reader)
    labels = {}

    # WHEN
    async def run_test():
        async with app.run_test() as pilot:
            await pilot.pause()
            labels.update(extract_label_text(pilot.app))

    async_run(run_test())

    # THEN
    assert labels["cmd"].rstrip() == display_val


def test_header_with_no_snapshots():
    # GIVEN
    reader = MockReader([])
    app = MockApp(reader)
    labels = {}

    # WHEN
    async def run_test():
        async with app.run_test() as pilot:
            await pilot.pause()
            labels.update(extract_label_text(pilot.app))

    async_run(run_test())

    # THEN
    assert labels["tid"].split() == "TID: *".split()
    assert labels["thread"].split() == "All threads".split()
    assert labels["samples"].split() == "Samples: 0".split()


def test_header_with_empty_snapshot():
    # GIVEN
    reader = MockReader([])
    app = MockApp(reader)
    labels = {}

    # WHEN
    async def run_test():
        async with app.run_test() as pilot:
            app.add_mock_snapshot([])
            await pilot.pause()
            labels.update(extract_label_text(pilot.app))

    async_run(run_test())

    # THEN
    assert labels["tid"].split() == "TID: *".split()
    assert labels["thread"].split() == "All threads".split()
    assert labels["samples"].split() == "Samples: 1".split()


def test_sorting():
    """Test that our sort keys correctly sort the data table"""
    # GIVEN
    snapshot = [
        mock_allocation(
            size=10,
            n_allocations=5,
            stack=[("a", "a.py", 1)],
        ),
        mock_allocation(
            size=50,
            n_allocations=1,
            stack=[("b", "b.py", 1)],
        ),
        mock_allocation(
            size=100,
            n_allocations=2,
            stack=[("c", "c.py", 1), ("b", "b.py", 1)],
        ),
        mock_allocation(
            size=25,
            n_allocations=4,
            stack=[("d", "d.py", 1)],
        ),
    ]

    own_order = "cbda"
    total_order = "bcda"
    allocations_order = "adbc"

    reader = MockReader([])
    app = MockApp(reader)
    order_by_key = {}

    # WHEN
    async def run_test():
        async with app.run_test() as pilot:
            app.add_mock_snapshot(snapshot)
            await pilot.pause()

            datatable = pilot.app.screen.query_one(DataTable)
            function_col_key = datatable.ordered_columns[0].key

            for key in ("", "o", "a", "t"):
                await pilot.press(key)
                order_by_key[key] = "".join(
                    datatable.get_cell(row.key, function_col_key).plain
                    for row in datatable.ordered_rows
                )

    async_run(run_test())

    # THEN
    assert order_by_key[""] == total_order
    assert order_by_key["o"] == own_order
    assert order_by_key["a"] == allocations_order
    assert order_by_key["t"] == total_order


def test_switching_threads():
    """Test that we can switch which thread is displayed"""
    # GIVEN
    thread_names = ["Thread A", "", "Thread C"]
    thread_labels = [
        "Thread 1 of 3 (Thread A)",
        "Thread 2 of 3",
        "Thread 3 of 3 (Thread C)",
    ]
    snapshot = [
        mock_allocation(
            tid=1,
            stack=[("a", "a.py", 1)],
            thread_name=thread_names[0],
        ),
        mock_allocation(
            tid=2,
            stack=[("b", "b.py", 1)],
            thread_name=thread_names[1],
        ),
        mock_allocation(
            tid=3,
            stack=[("c", "c.py", 1)],
            thread_name=thread_names[2],
        ),
    ]

    reader = MockReader([])
    app = MockApp(reader)
    functions = []
    tids = []
    threads = []

    # WHEN
    async def run_test():
        async with app.run_test() as pilot:
            app.add_mock_snapshot(snapshot)
            await pilot.pause()

            datatable = pilot.app.screen.query_one(DataTable)

            for key in ("m", ">", ">", ">", "<", "<", "<"):
                await pilot.press(key)
                functions.append(datatable.get_cell_at(Coordinate(0, 0)).plain)
                labels = extract_label_text(app)
                tids.append(" ".join(labels["tid"].split()))
                threads.append(" ".join(labels["thread"].split()))

    async_run(run_test())

    # THEN
    order = [0, 1, 2, 0, 2, 1, 0]
    assert functions == ["abc"[i] for i in order]
    assert tids == [f"TID: {hex(i+1)}" for i in order]
    assert threads == [thread_labels[i] for i in order]


def test_merge_mode_new_threads():
    """Test that the 'All threads' is still displayed when a new thread is created."""
    # GIVEN
    snapshot = [
        mock_allocation(
            tid=1,
            stack=[("a", "a.py", 1)],
        ),
        mock_allocation(
            tid=2,
            stack=[("b", "b.py", 1)],
        ),
        mock_allocation(
            tid=3,
            stack=[("c", "c.py", 1)],
        ),
    ]
    new_thread = mock_allocation(tid=4, stack=[("d", "d.py", 1)])

    reader = MockReader([])
    app = MockApp(reader)
    label = []

    # WHEN
    async def run_test():
        async with app.run_test() as pilot:
            await pilot.press("m")
            app.add_mock_snapshot(snapshot)
            await pilot.pause()

            await pilot.press("m")
            app.add_mock_snapshot(snapshot + [new_thread])
            await pilot.pause()
            label.append(extract_label_text(app)["thread"])

    async_run(run_test())

    # THEN
    assert label == ["All threads\n"]


def test_merging_allocations_from_all_threads():
    """Test that we can display allocations from all threads"""
    # GIVEN
    snapshot = [
        mock_allocation(
            tid=1,
            size=1024,
            stack=[("a", "a.py", 1)],
        ),
        mock_allocation(
            tid=2,
            size=2 * 1024,
            stack=[("b", "b.py", 1)],
        ),
        mock_allocation(
            tid=3,
            size=3 * 1024,
            stack=[("c", "c.py", 1)],
        ),
    ]

    reader = MockReader([])
    app = MockApp(reader)
    functions = []
    tids = []
    threads = []

    # WHEN
    async def run_test():
        async with app.run_test() as pilot:
            app.add_mock_snapshot(snapshot)
            await pilot.pause()

            datatable = pilot.app.screen.query_one(DataTable)

            for key in ("m", ">", "m", "<", "m", "<"):
                await pilot.press(key)
                functions.append(datatable.get_cell_at(Coordinate(0, 0)).plain)
                labels = extract_label_text(app)
                tids.append(" ".join(labels["tid"].split()))
                threads.append(" ".join(labels["thread"].split()))

    async_run(run_test())

    # THEN
    order = [0, 1, 2, 2, 1, 0]
    merged = [False, False, True, True, False, False]
    assert functions == ["abc"[i] for i in order]
    assert tids == [
        "TID: *" if all else f"TID: {hex(i+1)}" for i, all in zip(order, merged)
    ]
    assert threads == [
        "All threads" if all else f"Thread {i+1} of 3" for i, all in zip(order, merged)
    ]


@pytest.mark.parametrize(
    "terminal_size, press, snapshots",
    [
        pytest.param(
            (80, 24), [], SHORT_SNAPSHOTS, id="narrow-terminal-short-snapshots"
        ),
        pytest.param(
            (80, 24),
            ["tab"],
            LONG_SNAPSHOTS,
            id="narrow-terminal-focus-header-long-snapshots",
        ),
        pytest.param((120, 24), [], LONG_SNAPSHOTS, id="wide-terminal-long-snapshots"),
        pytest.param(
            (200, 24), [], SHORT_SNAPSHOTS, id="very-wide-terminal-short-snapshots"
        ),
    ],
)
def test_tui_basic(terminal_size, press, snapshots, compare):
    async def run_before(pilot) -> None:
        pilot.app.add_mock_snapshots(snapshots)

    assert compare(
        press=press,
        run_before=run_before,
        terminal_size=terminal_size,
    )


@pytest.mark.parametrize(
    "terminal_size, disconnected",
    [
        pytest.param((50, 24), False, id="narrow-terminal-connected"),
        pytest.param((50, 24), True, id="narrow-terminal-disconnected"),
        pytest.param((81, 24), True, id="wider-terminal"),
    ],
)
def test_tui_pause(terminal_size, disconnected, compare):
    async def run_before(pilot: Pilot) -> None:
        app = cast(MockApp, pilot.app)
        app.add_mock_snapshot(SHORT_SNAPSHOTS[0])
        await pilot.pause()
        await pilot.press("space")
        await pilot.press("tab")
        await pilot.pause()
        app.add_mock_snapshot(SHORT_SNAPSHOTS[1], disconnected=disconnected)

    assert compare(
        run_before=run_before,
        terminal_size=terminal_size,
    )


def test_tui_gradient(compare):
    snapshot = [
        mock_allocation(
            stack=[(f"function{j}", f"/abc/lel_{j}.py", i) for j in range(i, -1, -1)],
            size=1024 + 10 * i,
            n_allocations=1,
        )
        for i in range(0, 30)
    ]

    async def run_before(pilot) -> None:
        pilot.app.add_mock_snapshots([snapshot], native=False)

    assert compare(run_before=run_before, terminal_size=(125, 40), native=False)


class TestAggregateResults:
    def test_simple_allocations(self):
        # GIVEN
        mock_allocation_records = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=10,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=2,
                _stack=[
                    ("me", "fun.py", 12),
                    ("parent", "fun.py", 8),
                    ("grandparent", "fun.py", 4),
                ],
            ),
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=20,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _stack=[
                    ("sibling", "fun.py", 16),
                    ("parent", "fun.py", 8),
                    ("grandparent", "fun.py", 4),
                ],
            ),
        ]
        allocation_records = cast(List[AllocationRecord], mock_allocation_records)

        # WHEN
        result = aggregate_allocations(allocation_records)

        # THEN
        grandparent = result[Location(function="grandparent", file="fun.py")]
        assert grandparent.own_memory == 0
        assert grandparent.total_memory == 30
        assert grandparent.n_allocations == 3

        me = result[Location(function="me", file="fun.py")]
        assert me.own_memory == 10
        assert me.total_memory == 10
        assert me.n_allocations == 2

        parent = result[Location(function="parent", file="fun.py")]
        assert parent.own_memory == 0
        assert parent.total_memory == 30
        assert parent.n_allocations == 3

    def test_missing_frames(self):
        # GIVEN
        mock_allocation_records = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=10,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=2,
                _stack=[],
            ),
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=20,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _stack=[
                    ("sibling", "fun.py", 16),
                    ("parent", "fun.py", 8),
                    ("grandparent", "fun.py", 4),
                ],
            ),
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=30,
                allocator=AllocatorType.MALLOC,
                stack_id=2,
                n_allocations=1,
                _stack=[],
            ),
        ]
        allocation_records = cast(List[AllocationRecord], mock_allocation_records)

        # WHEN
        result = aggregate_allocations(allocation_records)

        # THEN
        grandparent = result[Location(function="grandparent", file="fun.py")]
        assert grandparent.own_memory == 0
        assert grandparent.total_memory == 20
        assert grandparent.n_allocations == 1

        me = result[Location(function="???", file="???")]
        assert me.own_memory == 40
        assert me.total_memory == 40
        assert me.n_allocations == 3

    def test_native_frames(self):
        # GIVEN
        mock_allocation_records = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=10,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=2,
                _stack=[],
                _hybrid_stack=[],
            ),
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=20,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _hybrid_stack=[
                    ("sibling", "fun.c", 16),
                    ("parent", "fun.py", 8),
                    ("grandparent", "fun.py", 4),
                ],
            ),
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=30,
                allocator=AllocatorType.MALLOC,
                stack_id=2,
                n_allocations=1,
                _hybrid_stack=[],
            ),
        ]
        allocation_records = cast(List[AllocationRecord], mock_allocation_records)

        # WHEN
        result = aggregate_allocations(allocation_records, native_traces=True)

        # THEN
        grandparent = result[Location(function="grandparent", file="fun.py")]
        assert grandparent.own_memory == 0
        assert grandparent.total_memory == 20
        assert grandparent.n_allocations == 1

        me = result[Location(function="???", file="???")]
        assert me.own_memory == 40
        assert me.total_memory == 40
        assert me.n_allocations == 3


def test_merge_threads(compare):
    async def run_before(pilot: Pilot) -> None:
        snapshot = [
            mock_allocation(
                tid=1,
                stack=[("a", "a.py", 1)],
            ),
            mock_allocation(
                tid=2,
                stack=[("b", "b.py", 1)],
            ),
            mock_allocation(
                tid=3,
                stack=[("c", "c.py", 1)],
            ),
        ]
        app = cast(MockApp, pilot.app)
        await pilot.press("m")
        app.add_mock_snapshot(snapshot)
        await pilot.pause()
        await pilot.press("m")
        await pilot.pause()
        app.add_mock_snapshot(snapshot)

    assert compare(
        run_before=run_before,
        terminal_size=(150, 24),
    )


def test_unmerge_threads(compare):
    async def run_before(pilot: Pilot) -> None:
        snapshot = [
            mock_allocation(
                tid=1,
                stack=[("a", "a.py", 1)],
            ),
            mock_allocation(
                tid=2,
                stack=[("b", "b.py", 1)],
            ),
            mock_allocation(
                tid=3,
                stack=[("c", "c.py", 1)],
            ),
        ]
        app = cast(MockApp, pilot.app)
        app.add_mock_snapshot(snapshot)
        await pilot.press("m")
        await pilot.pause()
        await pilot.press(">")
        await pilot.press("m")
        await pilot.press(">")
        await pilot.press("m")
        await pilot.pause()
        app.add_mock_snapshot(snapshot)

    assert compare(
        run_before=run_before,
        terminal_size=(150, 24),
    )



from memray import AllocatorType
from memray.reporters.table import TableReporter
from tests.utils import MockAllocationRecord


class TestTableReporter:
    def test_empty_report(self):
        # GIVEN / WHEN
        table = TableReporter.from_snapshot([], memory_records=[], native_traces=False)

        # THEN
        assert table.data == []

    def test_single_allocation(self):
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=1024,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _stack=[
                    ("me", "fun.py", 12),
                ],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations, memory_records=[], native_traces=False
        )

        # THEN
        assert table.data == [
            {
                "tid": "0x1",
                "size": 1024,
                "allocator": "malloc",
                "n_allocations": 1,
                "stack_trace": "me at fun.py:12",
            }
        ]

    def test_single_native_allocation(self):
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=1024,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _hybrid_stack=[
                    ("me", "fun.c", 12),
                ],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations, memory_records=[], native_traces=True
        )

        # THEN
        assert table.data == [
            {
                "tid": "0x1",
                "size": 1024,
                "allocator": "malloc",
                "n_allocations": 1,
                "stack_trace": "me at fun.c:12",
            }
        ]

    def test_multiple_allocations(self):
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=1024,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _stack=[
                    ("me", "foo.py", 12),
                ],
            ),
            MockAllocationRecord(
                tid=1,
                address=0x1100000,
                size=2048,
                allocator=AllocatorType.VALLOC,
                stack_id=2,
                n_allocations=10,
                _stack=[
                    ("you", "bar.py", 21),
                ],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations, memory_records=[], native_traces=False
        )

        # THEN
        assert table.data == [
            {
                "tid": "0x1",
                "size": 1024,
                "allocator": "malloc",
                "n_allocations": 1,
                "stack_trace": "me at foo.py:12",
            },
            {
                "tid": "0x1",
                "size": 2048,
                "allocator": "valloc",
                "n_allocations": 10,
                "stack_trace": "you at bar.py:21",
            },
        ]

    def test_aggregates_records_with_same_top_frame(self):
        """Records with different full stacks but same top frame should be aggregated."""
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=1024,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _stack=[
                    ("me", "fun.py", 12),
                    ("caller_a", "a.py", 1),
                ],
            ),
            MockAllocationRecord(
                tid=1,
                address=0x1100000,
                size=2048,
                allocator=AllocatorType.MALLOC,
                stack_id=2,
                n_allocations=3,
                _stack=[
                    ("me", "fun.py", 12),
                    ("caller_b", "b.py", 5),
                ],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations, memory_records=[], native_traces=False
        )

        # THEN
        assert table.data == [
            {
                "tid": "0x1",
                "size": 1024 + 2048,
                "allocator": "malloc",
                "n_allocations": 4,
                "stack_trace": "me at fun.py:12",
            }
        ]

    def test_does_not_aggregate_different_allocators(self):
        """Records with the same top frame but different allocators stay separate."""
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=1024,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _stack=[
                    ("me", "fun.py", 12),
                ],
            ),
            MockAllocationRecord(
                tid=1,
                address=0x1100000,
                size=2048,
                allocator=AllocatorType.CALLOC,
                stack_id=2,
                n_allocations=1,
                _stack=[
                    ("me", "fun.py", 12),
                ],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations, memory_records=[], native_traces=False
        )

        # THEN
        assert len(table.data) == 2
        assert table.data[0]["allocator"] == "malloc"
        assert table.data[1]["allocator"] == "calloc"

    def test_does_not_aggregate_different_threads_when_split(self):
        """With --split-threads, records from different threads stay separate."""
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=1024,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _stack=[
                    ("me", "fun.py", 12),
                ],
            ),
            MockAllocationRecord(
                tid=2,
                address=0x1100000,
                size=2048,
                allocator=AllocatorType.MALLOC,
                stack_id=2,
                n_allocations=1,
                _stack=[
                    ("me", "fun.py", 12),
                ],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations,
            memory_records=[],
            native_traces=False,
            merge_threads=False,
        )

        # THEN
        assert len(table.data) == 2
        assert table.data[0]["tid"] == "0x1"
        assert table.data[1]["tid"] == "0x2"

    def test_aggregates_different_threads_when_merged(self):
        """With merged threads (the default), records with the same displayed
        fields are combined even if they come from different threads, since the
        thread ID column is hidden."""
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=1024,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _stack=[
                    ("me", "fun.py", 12),
                ],
            ),
            MockAllocationRecord(
                tid=2,
                address=0x1100000,
                size=2048,
                allocator=AllocatorType.MALLOC,
                stack_id=2,
                n_allocations=3,
                _stack=[
                    ("me", "fun.py", 12),
                ],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations, memory_records=[], native_traces=False
        )

        # THEN
        assert table.data == [
            {
                "tid": "0x1",
                "size": 1024 + 2048,
                "allocator": "malloc",
                "n_allocations": 4,
                "stack_trace": "me at fun.py:12",
            }
        ]

    def test_aggregates_native_records_with_same_top_frame(self):
        """Native trace records with same top frame should also be aggregated."""
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=512,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=2,
                _hybrid_stack=[
                    ("alloc", "mem.c", 42),
                    ("caller_a", "a.c", 10),
                ],
            ),
            MockAllocationRecord(
                tid=1,
                address=0x1100000,
                size=256,
                allocator=AllocatorType.MALLOC,
                stack_id=2,
                n_allocations=5,
                _hybrid_stack=[
                    ("alloc", "mem.c", 42),
                    ("caller_b", "b.c", 20),
                ],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations, memory_records=[], native_traces=True
        )

        # THEN
        assert table.data == [
            {
                "tid": "0x1",
                "size": 768,
                "allocator": "malloc",
                "n_allocations": 7,
                "stack_trace": "alloc at mem.c:42",
            }
        ]

    def test_html_special_chars_in_stack_trace(self):
        """Stack traces with HTML special chars are escaped in output."""
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=512,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _hybrid_stack=[
                    ("std::vector<int>::push_back", "vector.h", 100),
                ],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations, memory_records=[], native_traces=True
        )

        # THEN
        assert len(table.data) == 1
        assert "&lt;" in table.data[0]["stack_trace"]
        assert "&gt;" in table.data[0]["stack_trace"]
        assert "<int>" not in table.data[0]["stack_trace"]

    def test_html_special_chars_do_not_break_aggregation(self):
        """Records with HTML chars in function names still aggregate correctly."""
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=512,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _hybrid_stack=[
                    ("std::vector<int>::push_back", "vector.h", 100),
                    ("caller_a", "a.cpp", 10),
                ],
            ),
            MockAllocationRecord(
                tid=1,
                address=0x1100000,
                size=256,
                allocator=AllocatorType.MALLOC,
                stack_id=2,
                n_allocations=2,
                _hybrid_stack=[
                    ("std::vector<int>::push_back", "vector.h", 100),
                    ("caller_b", "b.cpp", 20),
                ],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations, memory_records=[], native_traces=True
        )

        # THEN
        assert len(table.data) == 1
        assert table.data[0]["size"] == 768
        assert table.data[0]["n_allocations"] == 3

    def test_empty_stack_trace(self):
        # GIVEN
        peak_allocations = [
            MockAllocationRecord(
                tid=1,
                address=0x1000000,
                size=1024,
                allocator=AllocatorType.MALLOC,
                stack_id=1,
                n_allocations=1,
                _stack=[],
            ),
        ]

        # WHEN
        table = TableReporter.from_snapshot(
            peak_allocations, memory_records=[], native_traces=False
        )

        # THEN
        assert table.data == [
            {
                "tid": "0x1",
                "size": 1024,
                "allocator": "malloc",
                "n_allocations": 1,
                "stack_trace": "???",
            }
        ]



import json
from collections import Counter
from datetime import datetime
from typing import List
from typing import Optional
from typing import Tuple
from unittest.mock import patch

import pytest

from memray import AllocatorType as AT
from memray._memray import FileFormat
from memray._metadata import Metadata
from memray._stats import Stats
from memray.reporters.stats import StatsReporter
from memray.reporters.stats import draw_histogram
from memray.reporters.stats import get_histogram_databins
from tests.utils import MockAllocationRecord


# data generator for tests
def _generate_mock_allocations(
    count: int,
    sizes: Optional[List[int]] = None,
    allocators: Optional[List[AT]] = None,
    n_allocations: Optional[List[int]] = None,
    stacks: Optional[List[List[Tuple[str, str, int]]]] = None,
):  # pragma: no cover
    if sizes is None:
        sizes = []
    if allocators is None:
        allocators = []
    if n_allocations is None:
        n_allocations = []
    if stacks is None:
        stacks = []

    sizes.extend([1024] * (count - len(sizes)))
    sizes = sizes[:count]

    allocators.extend([AT.MALLOC] * (count - len(allocators)))
    allocators = allocators[:count]

    n_allocations.extend([1] * (count - len(n_allocations)))
    n_allocations = n_allocations[:count]

    default_stacks_value = [
        ("me", "fun.py", 12),
        ("parent", "fun.py", 8),
        ("grandparent", "fun.py", 4),
    ]
    stacks.extend([default_stacks_value] * (count - len(stacks)))
    stacks = stacks[:count]

    return [
        MockAllocationRecord(
            tid=i + 1,
            address=0x1000000,
            size=sizes[i],
            allocator=allocators[i],
            stack_id=i + 1,
            n_allocations=n_allocations[i],
            _stack=stacks[i],
        )
        for i in range(count)
    ]


# data generator for tests
@pytest.fixture(scope="module")
def fake_stats():
    mem_allocation_list = [
        2500,
        11000,
        11000,
        12000,
        60000,
        65000,
        120000,
        125000,
        125000,
        160000,
        170000,
        180000,
        800000,
        1500000,
    ]

    s = Stats(
        metadata=Metadata(
            start_time=datetime(2023, 1, 1, 1),
            end_time=datetime(2023, 1, 1, 2),
            total_allocations=sum(mem_allocation_list),
            total_frames=10,
            peak_memory=max(mem_allocation_list),
            command_line="fake stats",
            pid=123456,
            python_allocator="pymalloc",
            has_native_traces=False,
            trace_python_allocators=True,
            file_format=FileFormat.ALL_ALLOCATIONS,
            main_thread_id=0x1,
        ),
        total_num_allocations=20,
        total_memory_allocated=sum(mem_allocation_list),
        peak_memory_allocated=max(mem_allocation_list),
        allocation_count_by_size=Counter(mem_allocation_list),
        allocation_count_by_allocator={
            AT.MALLOC.name: 1013,
            AT.REALLOC.name: 797,
            AT.CALLOC.name: 152,
            AT.MMAP.name: 4,
        },
        top_locations_by_count=[
            (("fake_func", "fake.py", 5), 20),
            (("fake_func2", "fake.py", 10), 50),
            (("__main__", "fake.py", 15), 1),
        ],
        top_locations_by_size=[
            (("fake_func", "fake.py", 5), 5 * 2**20),
            (("fake_func2", "fake.py", 10), 3 * 2**10),
            (("__main__", "fake.py", 15), 4),
        ],
        top_modules_by_allocation_size=[
            ("fake", 20, sum(mem_allocation_list)),
        ],
        top_modules_by_allocation_count=[
            ("fake", 20, sum(mem_allocation_list)),
        ],
    )
    return s


@pytest.fixture(scope="module")
def empty_module_stats():
    """Stats with no module attribution (top_modules_by_allocation_size is empty)."""
    return Stats(
        metadata=Metadata(
            start_time=datetime(2023, 1, 1, 1),
            end_time=datetime(2023, 1, 1, 2),
            total_allocations=20,
            total_frames=4,
            peak_memory=2500000,
            command_line="python -c 'import json; json.dumps(range(100))'",
            pid=123456,
            python_allocator="pymalloc",
            has_native_traces=False,
            trace_python_allocators=True,
            file_format=FileFormat.ALL_ALLOCATIONS,
            main_thread_id=0x1,
        ),
        total_num_allocations=20,
        total_memory_allocated=3341500,
        peak_memory_allocated=2500000,
        allocation_count_by_size=Counter({256: 12, 512: 8}),
        allocation_count_by_allocator={
            AT.MALLOC.name: 15,
            AT.REALLOC.name: 5,
        },
        top_locations_by_count=[
            (("loads", "/usr/lib/python3.11/json/__init__.py", 346), 12),
            (("decode", "/usr/lib/python3.11/json/decoder.py", 337), 8),
        ],
        top_locations_by_size=[
            (("loads", "/usr/lib/python3.11/json/__init__.py", 346), 2500000),
            (("decode", "/usr/lib/python3.11/json/decoder.py", 337), 841500),
        ],
        top_modules_by_allocation_size=[],
        top_modules_by_allocation_count=[],
    )


# tests begin here
def test_get_histogram_databins():
    # GIVEN
    input_data = Counter(
        [
            2500,
            11000,
            11000,
            12000,
            60000,
            65000,
            120000,
            125000,
            125000,
            160000,
            170000,
            180000,
            800000,
            1500000,
        ]
    )
    expected_output = [
        (8986, 1),
        (32299, 3),
        (116099, 2),
        (417312, 6),
        (1500000, 2),
    ]

    # WHEN
    actual_output = get_histogram_databins(input_data, bins=5)

    # THEN
    assert expected_output == actual_output


def test_get_histogram_databins_rounding():
    """Data chosen to provoke a floating point rounding error.

    In particular, so that:

        log(low) + sum([(log(high) - log(low)) / bins] * bins) > log(high)
    """
    # GIVEN
    input_data = Counter(
        [
            32,
            1050856,
        ]
    )
    expected_output = [
        (90, 1),
        (256, 0),
        (724, 0),
        (2049, 0),
        (5798, 0),
        (16405, 0),
        (46411, 0),
        (131299, 0),
        (371453, 0),
        (1050856, 1),
    ]

    # WHEN
    actual_output = get_histogram_databins(input_data, bins=10)

    # THEN
    assert expected_output == actual_output


def test_get_histogram_over_bound():
    """Data chosen to provoke a scenario where the computed allocation exceeds the upper limit.

    In particular, so that:
        Counter(min((x - low) // step, bins-1) for x in it) will default to placing it in the
        last bin instead of creating a new record out of range of the bins.
    """
    input_data = Counter([10000000000, 536, 536, 592, 576, 4486])
    expected_output = [
        (2859, 4),
        (15252, 1),
        (81360, 0),
        (434009, 0),
        (2315167, 0),
        (12349970, 0),
        (65879369, 0),
        (351425246, 0),
        (1874633954, 0),
        (10000000000, 1),
    ]

    # WHEN
    actual_output = get_histogram_databins(input_data, bins=10)

    # THEN
    assert expected_output == actual_output


def test_get_histogram_all_allocations_same_size():
    input_data = Counter([10000000000, 10000000000, 10000000000])
    expected_output = [
        (316227, 0),
        (999999, 0),
        (3162277, 0),
        (10000000, 0),
        (31622776, 0),
        (100000000, 0),
        (316227766, 0),
        (999999999, 0),
        (3162277660, 0),
        (10000000000, 3),
    ]

    # WHEN
    actual_output = get_histogram_databins(input_data, bins=10)

    # THEN
    assert expected_output == actual_output


def test_get_histogram_databins_invalid_bins():
    with pytest.raises(ValueError):
        _ = get_histogram_databins([], bins=0)  # invalid bins value
    with pytest.raises(ValueError):
        _ = get_histogram_databins([], bins=-1)  # invalid bins value


def test_draw_histogram():
    # GIVEN
    input_data = Counter(
        [
            2500,
            11000,
            11000,
            12000,
            60000,
            65000,
            120000,
            125000,
            125000,
            160000,
            170000,
            180000,
            800000,
            1500000,
        ]
    )
    expected_output = """min: 2.500kB
\t----------------------------------------
\t< 8.986kB  : 1 ▇▇▇▇▇
\t< 32.299kB : 3 ▇▇▇▇▇▇▇▇▇▇▇▇▇
\t< 116.099kB: 2 ▇▇▇▇▇▇▇▇▇
\t< 417.312kB: 6 ▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇
\t<=1.500MB  : 2 ▇▇▇▇▇▇▇▇▇
\t----------------------------------------
\tmax: 1.500MB"""

    # WHEN
    actual_output = draw_histogram(input_data, bins=5)

    # THEN
    assert expected_output == actual_output


def test_draw_histogram_smaller_scale_factor():
    # GIVEN
    input_data = Counter(
        [
            2500,
            11000,
            11000,
            12000,
            60000,
            65000,
            120000,
            125000,
            125000,
            160000,
            170000,
            180000,
            800000,
            1500000,
        ]
    )

    expected_output = """min: 2.500kB
\t--------------------
\t< 8.986kB  : 1 ▇
\t< 32.299kB : 3 ▇▇▇
\t< 116.099kB: 2 ▇▇
\t< 417.312kB: 6 ▇▇▇▇▇
\t<=1.500MB  : 2 ▇▇
\t--------------------
\tmax: 1.500MB"""

    # WHEN
    actual_output = draw_histogram(
        input_data, bins=5, hist_scale_factor=5
    )  # setting this to 5

    # THEN
    assert expected_output == actual_output


def test_draw_histogram_invalid_input():
    # test#1 - No input data
    input_data = Counter()
    actual_output = draw_histogram(input_data, bins=5)
    assert "<no data for histogram>" == actual_output

    # test#2 - Invalid bins value
    with pytest.raises(ValueError):
        _ = draw_histogram([100, 200, 300], bins=0)

    # test#3 - Invalid hist_scale_factor value
    with pytest.raises(ValueError):
        _ = draw_histogram([100, 200, 300], bins=5, hist_scale_factor=0)


def test_stats_output(fake_stats):
    reporter = StatsReporter(fake_stats, 5)
    with patch("builtins.print") as mocked_print:
        with patch("rich.print", print):
            reporter.render()
    expected = (
        "📏 [bold]Total allocations:[/]\n"
        "\t20\n"
        "\n"
        "📦 [bold]Total memory allocated:[/]\n"
        "\t3.341MB\n"
        "\n"
        "📈 [bold]Peak memory usage:[/]\n"
        "\t1.500MB\n"
        "\n"
        "📊 [bold]Histogram of allocation size:[/]\n"
        "\tmin: 2.500kB\n"
        "\t----------------------------------------\n"
        "\t< 4.739kB  : 1 ▇▇▇▇▇\n"
        "\t< 8.986kB  : 0 \n"
        "\t< 17.036kB : 3 ▇▇▇▇▇▇▇▇▇▇▇▇▇\n"
        "\t< 32.299kB : 0 \n"
        "\t< 61.237kB : 1 ▇▇▇▇▇\n"
        "\t< 116.099kB: 1 ▇▇▇▇▇\n"
        "\t< 220.113kB: 6 ▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇▇\n"
        "\t< 417.312kB: 0 \n"
        "\t< 791.181kB: 0 \n"
        "\t<=1.500MB  : 2 ▇▇▇▇▇▇▇▇▇\n"
        "\t----------------------------------------\n"
        "\tmax: 1.500MB\n"
        "\n"
        "📂 [bold]Allocator type distribution:[/]\n"
        "\t MALLOC: 1013\n"
        "\t REALLOC: 797\n"
        "\t CALLOC: 152\n"
        "\t MMAP: 4\n"
        "\n"
        "🥇 [bold]Top 5 largest allocating locations (by size):[/]\n"
        "\t- fake_func:fake.py:5 -> 5.243MB\n"
        "\t- fake_func2:fake.py:10 -> 3.072kB\n"
        "\t- __main__:fake.py:15 -> 4.000B\n"
        "\n"
        "🥇 [bold]Top 5 largest allocating locations (by number of allocations):[/]\n"
        "\t- fake_func:fake.py:5 -> 20\n"
        "\t- fake_func2:fake.py:10 -> 50\n"
        "\t- __main__:fake.py:15 -> 1\n"
        "\n"
        "🥇 [bold]Top 5 allocating modules (by size):[/]\n"
        "\t- fake: 3.341MB total | 20 allocations\n"
        "\n"
        "🥇 [bold]Top 5 allocating modules (by number of allocations):[/]\n"
        "\t- fake: 3.341MB total | 20 allocations"
    )
    printed = "\n".join(" ".join(x[0]) for x in mocked_print.call_args_list)
    assert expected == printed


def test_stats_output_json(fake_stats, tmp_path):
    output_file = tmp_path / "json.out"
    reporter = StatsReporter(fake_stats, 5)
    reporter.render(json_output_file=output_file)
    expected = {
        "total_num_allocations": 20,
        "total_bytes_allocated": 3341500,
        "allocation_size_histogram": [
            {"min_bytes": 0, "max_bytes": 4738, "count": 1},
            {"min_bytes": 4739, "max_bytes": 8985, "count": 0},
            {"min_bytes": 8986, "max_bytes": 17035, "count": 3},
            {"min_bytes": 17036, "max_bytes": 32298, "count": 0},
            {"min_bytes": 32299, "max_bytes": 61236, "count": 1},
            {"min_bytes": 61237, "max_bytes": 116098, "count": 1},
            {"min_bytes": 116099, "max_bytes": 220112, "count": 6},
            {"min_bytes": 220113, "max_bytes": 417311, "count": 0},
            {"min_bytes": 417312, "max_bytes": 791180, "count": 0},
            {"min_bytes": 791181, "max_bytes": 1500000, "count": 2},
        ],
        "allocator_type_distribution": {
            "MALLOC": 1013,
            "REALLOC": 797,
            "CALLOC": 152,
            "MMAP": 4,
        },
        "top_allocations_by_size": [
            {"location": "fake_func:fake.py:5", "size": 5242880},
            {"location": "fake_func2:fake.py:10", "size": 3072},
            {"location": "__main__:fake.py:15", "size": 4},
        ],
        "top_allocations_by_count": [
            {"location": "fake_func:fake.py:5", "count": 20},
            {"location": "fake_func2:fake.py:10", "count": 50},
            {"location": "__main__:fake.py:15", "count": 1},
        ],
        "top_modules_by_allocation_size": [
            {"module": "fake", "num_allocations": 20, "total_bytes": 3341500},
        ],
        "top_modules_by_allocation_count": [
            {"module": "fake", "num_allocations": 20, "total_bytes": 3341500},
        ],
        "metadata": {
            "start_time": "2023-01-01 01:00:00",
            "end_time": "2023-01-01 02:00:00",
            "total_allocations": 3341500,
            "total_frames": 10,
            "peak_memory": 1500000,
            "command_line": "fake stats",
            "pid": 123456,
            "python_allocator": "pymalloc",
            "has_native_traces": False,
            "trace_python_allocators": True,
            "file_format": 0,
            "main_thread_id": 0x1,
        },
    }
    actual = json.loads(output_file.read_text())
    assert expected == actual


def test_stats_terminal_skips_empty_modules(empty_module_stats):
    reporter = StatsReporter(empty_module_stats, 5)
    with patch("builtins.print") as mocked_print:
        with patch("rich.print", print):
            reporter.render()
    printed = "\n".join(" ".join(x[0]) for x in mocked_print.call_args_list)
    assert "allocating modules" not in printed


def test_stats_json_includes_empty_module_list(empty_module_stats, tmp_path):
    output_file = tmp_path / "out.json"
    reporter = StatsReporter(empty_module_stats, 5)
    reporter.render(json_output_file=output_file)
    data = json.loads(output_file.read_text())
    assert data["top_modules_by_allocation_size"] == []
    assert data["top_modules_by_allocation_count"] == []



import json
import sys


EMPTY = ({}, 2000)
SIMPLE_DATA = {
    "key1": 0,
    "key2": True,
    "key3": "value",
    "key4": "foo",
    "key5": "string",
}
SIMPLE = (SIMPLE_DATA, 1000)
NESTED_DATA = {
    "key1": 0,
    "key2": SIMPLE[0],
    "key3": "value",
    "key4": SIMPLE[0],
    "key5": SIMPLE[0],
    "key": "\u0105\u0107\u017c",
}
NESTED = (NESTED_DATA, 1000)
HUGE = ([NESTED[0]] * 1000, 1)

CASES = ["EMPTY", "SIMPLE", "NESTED", "HUGE"]


def bench_json_dumps(data):
    for obj, count_it in data:
        for _ in count_it:
            json.dumps(obj)


def add_cmdline_args(cmd, args):
    if args.cases:
        cmd.extend(("--cases", args.cases))


def run_benchmark():
    data = []
    for case in CASES:
        obj, count = globals()[case]
        data.append((obj, range(count)))
    bench_json_dumps(data)


def main():
    run_benchmark()
